# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import logging
from collections import defaultdict
from typing import Dict, List, Optional

import xoscar as mo

from ....lib.aio import alru_cache
from ...meta.api import MetaAPI
from ...storage.api import StorageAPI
from ..errors import TileableNotTracked

logger = logging.getLogger(__name__)


class LifecycleTrackerActor(mo.Actor):
    _meta_api: MetaAPI

    def __init__(self, session_id: str):
        self._session_id = session_id
        self._tileable_key_to_chunk_keys = dict()
        self._tileable_ref_counts = defaultdict(lambda: 0)
        self._chunk_ref_counts = defaultdict(lambda: 0)

        self._meta_api: Optional[MetaAPI] = None

    async def __post_create__(self):
        self._meta_api = await MetaAPI.create(self._session_id, self.address)

    async def __pre_destroy__(self):
        chunk_keys = [
            chunk_key
            for chunk_key, ref_count in self._chunk_ref_counts.items()
            if ref_count > 0
        ]
        # remove all chunks
        await self._remove_chunks(chunk_keys)

    @alru_cache
    async def _get_task_api(self):
        from ...task.api import TaskAPI

        return await TaskAPI.create(self._session_id, self.address)

    @staticmethod
    def gen_uid(session_id):
        return f"{session_id}_lifecycle_tracker"

    def _track(self, tileable_key: str, chunk_keys: List[str]):
        if tileable_key not in self._tileable_key_to_chunk_keys:
            self._tileable_key_to_chunk_keys[tileable_key] = []
        chunk_keys_set = set(self._tileable_key_to_chunk_keys[tileable_key])
        incref_chunk_keys = []
        tileable_ref_count = self._tileable_ref_counts.get(tileable_key, 0)
        for chunk_key in chunk_keys:
            if chunk_key in chunk_keys_set:
                continue
            if tileable_ref_count > 0:
                incref_chunk_keys.extend([chunk_key] * tileable_ref_count)
            self._tileable_key_to_chunk_keys[tileable_key].append(chunk_key)
        if incref_chunk_keys:
            self._incref_chunks(incref_chunk_keys)

    @mo.extensible
    async def track(self, tileable_key: str, chunk_keys: List[str]):
        return await asyncio.to_thread(self._track, tileable_key, chunk_keys)

    @classmethod
    def _check_ref_counts(cls, keys: List[str], ref_counts: List[int]):
        if ref_counts is not None and len(keys) != len(ref_counts):
            raise ValueError(
                f"`ref_counts` should have same size as `keys`, expect {len(keys)}, got {len(ref_counts)}"
            )

    def _incref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        counts = counts if counts is not None else itertools.repeat(1)
        for chunk_key, count in zip(chunk_keys, counts):
            self._chunk_ref_counts[chunk_key] += count

    async def incref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        self._check_ref_counts(chunk_keys, counts)
        return await asyncio.to_thread(self._incref_chunks, chunk_keys, counts=counts)

    def _get_remove_chunk_keys(self, chunk_keys: List[str], counts: List[int] = None):
        to_remove_chunk_keys = []
        counts = counts if counts is not None else itertools.repeat(1)
        for chunk_key, count in zip(chunk_keys, counts):
            ref_count = self._chunk_ref_counts[chunk_key]
            ref_count -= count
            assert ref_count >= 0, f"chunk key {chunk_key} will have negative ref count"
            self._chunk_ref_counts[chunk_key] = ref_count
            if ref_count == 0:
                # remove
                to_remove_chunk_keys.append(chunk_key)
        return to_remove_chunk_keys

    async def decref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        self._check_ref_counts(chunk_keys, counts)
        to_remove_chunk_keys = await asyncio.to_thread(
            self._get_remove_chunk_keys, chunk_keys, counts=counts
        )
        # make _remove_chunks release actor lock so that multiple `decref_chunks` can run concurrently.
        yield self._remove_chunks(to_remove_chunk_keys)

    async def _remove_chunks(self, to_remove_chunk_keys: List[str]):
        # get meta
        logger.debug(
            "Remove chunks %.500s with a refcount of zero", to_remove_chunk_keys
        )
        get_metas = []
        for to_remove_chunk_key in to_remove_chunk_keys:
            get_metas.append(
                self._meta_api.get_chunk_meta.delay(
                    to_remove_chunk_key, fields=["bands"], error="ignore"
                )
            )
        metas = await self._meta_api.get_chunk_meta.batch(*get_metas)

        # filter chunks that not exist
        new_to_remove_chunk_keys = []
        new_metas = []
        for to_remove_chunk_key, meta in zip(to_remove_chunk_keys, metas):
            if meta is not None:
                new_to_remove_chunk_keys.append(to_remove_chunk_key)
                new_metas.append(meta)
        to_remove_chunk_keys = new_to_remove_chunk_keys
        metas = new_metas

        all_bands = [meta["bands"] for meta in metas]
        key_to_addresses = dict()
        for to_remove_chunk_key, bands in zip(to_remove_chunk_keys, all_bands):
            key_to_addresses[to_remove_chunk_key] = bands

        # remove data via storage API
        storage_api_to_deletes = defaultdict(list)
        for key, bands in key_to_addresses.items():
            for band in bands:
                # storage API is cached for same arguments
                storage_api = await StorageAPI.create(
                    self._session_id, band[0], band[1]
                )
                storage_api_to_deletes[storage_api].append(
                    storage_api.delete.delay(key, error="ignore")
                )
        await asyncio.gather(
            *[
                storage_api.delete.batch(*deletes)
                for storage_api, deletes in storage_api_to_deletes.items()
            ]
        )

        # delete meta
        delete_metas = []
        for to_remove_chunk_key in to_remove_chunk_keys:
            delete_metas.append(
                self._meta_api.del_chunk_meta.delay(to_remove_chunk_key)
            )
        await self._meta_api.del_chunk_meta.batch(*delete_metas)

    def get_chunk_ref_counts(self, chunk_keys: List[str]) -> List[int]:
        return [self._chunk_ref_counts[chunk_key] for chunk_key in chunk_keys]

    def get_all_chunk_ref_counts(self) -> Dict[str, int]:
        result = dict()
        for chunk_key, ref_count in self._chunk_ref_counts.items():
            if ref_count > 0:
                result[chunk_key] = ref_count
        return result

    def _incref_tileables(self, tileable_keys: List[str], counts: List[int] = None):
        counts = counts if counts is not None else itertools.repeat(1)
        for tileable_key, count in zip(tileable_keys, counts):
            if tileable_key not in self._tileable_key_to_chunk_keys:
                raise TileableNotTracked(f"tileable {tileable_key} not tracked before")
            self._tileable_ref_counts[tileable_key] += count
            incref_chunk_keys = self._tileable_key_to_chunk_keys[tileable_key]
            # incref chunks for this tileable
            logger.debug(
                "Incref chunks %.500s while increfing tileable %s",
                incref_chunk_keys,
                tileable_key,
            )
            chunk_counts = None if count == 1 else [count] * len(incref_chunk_keys)
            self._incref_chunks(incref_chunk_keys, counts=chunk_counts)

    async def incref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        self._check_ref_counts(tileable_keys, counts)
        return await asyncio.to_thread(
            self._incref_tileables, tileable_keys, counts=counts
        )

    def _get_decref_chunk_keys(
        self, tileable_keys: List[str], counts: List[int] = None
    ) -> Dict[str, int]:
        decref_chunk_keys = dict()
        counts = counts if counts is not None else itertools.repeat(1)
        for tileable_key, count in zip(tileable_keys, counts):
            if tileable_key not in self._tileable_key_to_chunk_keys:
                raise TileableNotTracked(f"tileable {tileable_key} not tracked before")
            self._tileable_ref_counts[tileable_key] -= count

            for chunk_key in self._tileable_key_to_chunk_keys[tileable_key]:
                if chunk_key not in decref_chunk_keys:
                    decref_chunk_keys[chunk_key] = count
                else:
                    decref_chunk_keys[chunk_key] += count
        logger.debug(
            "Decref chunks %.500s while decrefing tileables %s",
            decref_chunk_keys,
            tileable_keys,
        )
        return decref_chunk_keys

    async def decref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        self._check_ref_counts(tileable_keys, counts)
        decref_chunk_key_to_counts = await asyncio.to_thread(
            self._get_decref_chunk_keys, tileable_keys, counts=counts
        )
        to_remove_chunk_keys = await asyncio.to_thread(
            self._get_remove_chunk_keys,
            list(decref_chunk_key_to_counts),
            counts=list(decref_chunk_key_to_counts.values()),
        )
        to_remove_tileable_keys = await asyncio.to_thread(
            list, (key for key in tileable_keys if self._tileable_ref_counts[key] <= 0)
        )
        coros = []
        if to_remove_chunk_keys:
            coros.append(self._remove_chunks(to_remove_chunk_keys))
        if to_remove_tileable_keys:
            task_api = await self._get_task_api()
            coros.append(task_api.remove_tileables(to_remove_tileable_keys))
        if coros:
            # release actor lock
            yield asyncio.gather(*coros)

    def get_tileable_ref_counts(self, tileable_keys: List[str]) -> List[int]:
        return [
            self._tileable_ref_counts[tileable_key] for tileable_key in tileable_keys
        ]



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from typing import Dict, Type, Union

import pandas as pd


class DataFrameCustomGroupByAggMixin(ABC):
    @classmethod
    @abstractmethod
    def execute_map(cls, op, in_data: pd.DataFrame) -> Union[pd.DataFrame, pd.Series]:
        """
        Map stage implement.

        Parameters
        -------
        op : Any operand
            DataFrame operand.
        in_data : pd.DataFrame
            Input dataframe.

        Returns
        -------
            The result of op map stage.
        """

    @classmethod
    @abstractmethod
    def execute_combine(
        cls, op, in_data: pd.DataFrame
    ) -> Union[pd.DataFrame, pd.Series]:
        """
        Combine stage implement.

        Parameters
        ----------
        op : Any operand
            DataFrame operand.
        in_data : pd.Dataframe
            Input dataframe.

        Returns
        -------
            The result of op combine stage.
        """

    @classmethod
    @abstractmethod
    def execute_agg(cls, op, in_data: pd.DataFrame) -> Union[pd.DataFrame, pd.Series]:
        """
        Agg stage implement.

        Parameters
        ----------
        op : Any operand
            DataFrame operand.
        in_data : pd.Dataframe
            Input dataframe.

        Returns
        -------
            The result of op agg stage.
        """


custom_agg_functions: Dict[str, Type[DataFrameCustomGroupByAggMixin]] = {}


def register_custom_groupby_agg_func(method_name: str):
    def wrap(func_type: Type[DataFrameCustomGroupByAggMixin]):
        custom_agg_functions[method_name] = func_type
        return func_type

    return wrap



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
from abc import ABCMeta, abstractmethod

from numpy.linalg import LinAlgError
from sklearn.base import MultiOutputMixin
from sklearn.utils.validation import _deprecate_positional_args, check_is_fitted

from ... import execute
from ... import tensor as mt
from ...tensor.datasource import tensor as astensor
from ..base import BaseEstimator, ClassifierMixin, RegressorMixin
from ..preprocessing import normalize as f_normalize
from ..utils.validation import FLOAT_DTYPES, _check_sample_weight, check_array


def _preprocess_data(
    X,
    y,
    fit_intercept,
    normalize=False,
    copy=True,
    sample_weight=None,
    return_mean=False,
    check_input=True,
):
    """Center and scale data.

    Centers data to have mean zero along axis 0. If fit_intercept=False or if
    the X is a sparse matrix, no centering is done, but normalization can still
    be applied. The function returns the statistics necessary to reconstruct
    the input data, which are X_offset, y_offset, X_scale, such that the output

        X = (X - X_offset) / X_scale

    X_scale is the L2 norm of X - X_offset. If sample_weight is not None,
    then the weighted mean of X and y is zero, and not the mean itself. If
    return_mean=True, the mean, eventually weighted, is returned, independently
    of whether X was centered (option used for optimization with sparse data in
    coordinate_descend).

    This is here because nearly all linear models will want their data to be
    centered. This function also systematically makes y consistent with X.dtype
    """
    if isinstance(sample_weight, numbers.Number):
        sample_weight = None
    if sample_weight is not None:
        sample_weight = astensor(sample_weight)

    X = astensor(X)
    y = astensor(y, dtype=X.dtype)

    if check_input:
        X = check_array(X, copy=copy, accept_sparse=["csr", "csc"], dtype=FLOAT_DTYPES)
    elif copy:
        if X.issparse():
            X = X.copy()
        else:
            X = X.copy(order="K")

    if fit_intercept:
        if X.issparse():
            raise NotImplementedError("Does not support sparse input!")
        else:
            X_offset = mt.average(X, axis=0, weights=sample_weight)
            X = X - X_offset
            if normalize:
                X, X_scale = f_normalize(X, axis=0, copy=False, return_norm=True)
            else:
                X_scale = mt.ones(X.shape[1], dtype=X.dtype)
        y_offset = mt.average(y, axis=0, weights=sample_weight)
        y = y - y_offset
    else:
        if X.issparse():
            raise NotImplementedError("Does not support sparse input!")
        X_offset = mt.zeros(X.shape[1], dtype=X.dtype)
        X_scale = mt.ones(X.shape[1], dtype=X.dtype)
        if y.ndim == 1:
            y_offset = X.dtype.type(0)
        else:
            y_offset = mt.zeros(y.shape[1], dtype=X.dtype)

    return X, y, X_offset, y_offset, X_scale


def _rescale_data(X, y, sample_weight):
    """Rescale data sample-wise by square root of sample_weight.

    For many linear models, this enables easy support for sample_weight.

    Returns
    -------
    X_rescaled : {array-like, sparse matrix}

    y_rescaled : {array-like, sparse matrix}
    """
    n_samples = X.shape[0]
    sample_weight = mt.asarray(sample_weight)
    if sample_weight.ndim == 0:
        sample_weight = mt.full(n_samples, sample_weight, dtype=sample_weight.dtype)
    sample_weight = mt.sqrt(sample_weight)
    sw_matrix = mt.diag(sample_weight, sparse=True)
    X = mt.dot(sw_matrix, X)
    y = mt.dot(sw_matrix, y)
    return X, y


class LinearModel(BaseEstimator, metaclass=ABCMeta):
    """Base class for Linear Models"""

    @abstractmethod
    def fit(self, X, y, sample_weight=None):
        """Fit model."""

    def _decision_function(self, X):
        check_is_fitted(self)

        X = self._validate_data(
            X, y="no_validation", accept_sparse=["csr", "csc", "coo"], reset=False
        )
        return mt.dot(X, self.coef_.T) + self.intercept_

    def predict(self, X):
        """
        Predict using the linear model.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples,)
            Returns predicted values.
        """
        return self._decision_function(X)

    _preprocess_data = staticmethod(_preprocess_data)

    def _set_intercept(self, X_offset, y_offset, X_scale):
        """Set the intercept_"""
        if self.fit_intercept:
            self.coef_ = self.coef_ / X_scale
            self.intercept_ = y_offset - mt.dot(X_offset, self.coef_.T)
            execute(self.coef_, self.intercept_)
        else:
            self.intercept_ = mt.tensor(0.0)
            self.intercept_.execute()

    def _more_tags(self):  # noqa: R0201  # pylint: disable=no-self-use
        return {"requires_y": True}


class LinearRegression(MultiOutputMixin, RegressorMixin, LinearModel):
    """
    Ordinary least squares Linear Regression.

    LinearRegression fits a linear model with coefficients w = (w1, ..., wp)
    to minimize the residual sum of squares between the observed targets in
    the dataset, and the targets predicted by the linear approximation.

    Parameters
    ----------
    fit_intercept : bool, default=True
        Whether to calculate the intercept for this model. If set
        to False, no intercept will be used in calculations
        (i.e. data is expected to be centered).

    normalize : bool, default=False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`~sklearn.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : bool, default=True
        If True, X will be copied; else, it may be overwritten.

    positive : bool, default=False
        When set to ``True``, forces the coefficients to be positive. This
        option is only supported for dense arrays.

    Attributes
    ----------
    coef_ : array of shape (n_features, ) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
        If multiple targets are passed during the fit (y 2D), this
        is a 2D array of shape (n_targets, n_features), while if only
        one target is passed, this is a 1D array of length n_features.

    rank_ : int
        Rank of matrix `X`. Only available when `X` is dense.

    singular_ : array of shape (min(X, y),)
        Singular values of `X`. Only available when `X` is dense.

    intercept_ : float or array of shape (n_targets,)
        Independent term in the linear model. Set to 0.0 if
        `fit_intercept = False`.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    See Also
    --------
    Ridge : Ridge regression addresses some of the
        problems of Ordinary Least Squares by imposing a penalty on the
        size of the coefficients with l2 regularization.
    Lasso : The Lasso is a linear model that estimates
        sparse coefficients with l1 regularization.
    ElasticNet : Elastic-Net is a linear regression
        model trained with both l1 and l2 -norm regularization of the
        coefficients.
    """

    @_deprecate_positional_args
    def __init__(
        self,
        *,
        fit_intercept=True,
        normalize=False,
        copy_X=True,
        positive=False,
    ):
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.copy_X = copy_X
        self.positive = positive

    def fit(self, X, y, sample_weight=None):
        """
        Fit linear model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values. Will be cast to X's dtype if necessary.

        sample_weight : array-like of shape (n_samples,), default=None
            Individual weights for each sample.

        Returns
        -------
        self : object
            Fitted Estimator.
        """
        accept_sparse = False if self.positive else ["csr", "csc", "coo"]

        X, y = self._validate_data(
            X, y, accept_sparse=accept_sparse, y_numeric=True, multi_output=True
        )

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)

        X, y, X_offset, y_offset, X_scale = self._preprocess_data(
            X,
            y,
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            copy=self.copy_X,
            sample_weight=sample_weight,
            return_mean=True,
        )

        if sample_weight is not None:
            # Sample weight can be implemented via a simple rescaling.
            X, y = _rescale_data(X, y, sample_weight)

        if self.positive:
            # TODO: implement optimize.nnls first
            raise NotImplementedError("Does not support positive coefficients!")
        elif X.issparse():
            # TODO: implement sparse.linalg.lsqr first
            raise NotImplementedError("Does not support sparse input!")
        else:
            try:
                # In numpy:
                #   Mat mul does NOT always satisfy associative law
                # Tyipical mistake:
                #   (mt.linalg.inv(X.T @ X) @ (X.T @ y)).T
                self.coef_ = (mt.linalg.inv(X.T @ X) @ X.T @ y).T
                self.coef_.execute()
            except LinAlgError:
                # TODO: implement linalg.lstsq first
                raise NotImplementedError("Does not support singular matrix!")

        if y.ndim == 1:
            self.coef_ = mt.ravel(self.coef_)
            self.coef_.execute()
        self._set_intercept(X_offset, y_offset, X_scale)
        return self


class LinearClassifierMixin(ClassifierMixin):
    """Mixin for linear classifiers.

    Handles prediction for sparse and dense X.
    """

    def decision_function(self, X):
        """
        Predict confidence scores for samples.

        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)
            Confidence scores per (sample, class) combination. In the binary
            case, confidence score for self.classes_[1] where >0 means this
            class would be predicted.
        """
        check_is_fitted(self)

        X = check_array(X, accept_sparse="csr")

        n_features = self.coef_.shape[1]
        if X.shape[1] != n_features:
            raise ValueError(
                "X has %d features per sample; expecting %d" % (X.shape[1], n_features)
            )

        scores = mt.dot(X, self.coef_.T) + self.intercept_
        return scores

    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape [n_samples]
            Predicted class label per sample.
        """
        scores = self.decision_function(X)
        indices = scores.argmax(axis=1)
        return self.classes_[indices].execute()



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas

from ..core.adapter import (
    MarsGetAttrProxy,
    MarsPlotAccessor,
    from_mars,
    register_converter,
    to_mars,
)
from ..core.utils.docstring import attach_module_callable_docstring
from .mars_adapters.core import install_members


@register_converter(from_cls_list=[MarsPlotAccessor])
class PlotAccessor(MarsGetAttrProxy):
    def __init__(self, obj):
        super().__init__(MarsPlotAccessor(to_mars(obj)))

    def __call__(self, *args, **kwargs):  # pragma: no cover
        return from_mars(self._mars_obj(*to_mars(args), **to_mars(kwargs)))


install_members(
    PlotAccessor,
    MarsPlotAccessor,
    pandas,
    pandas.plotting._core.PlotAccessor,
)
attach_module_callable_docstring(
    PlotAccessor, pandas, pandas.plotting._core.PlotAccessor
)
PlotAccessor.__call__.__doc__ = PlotAccessor.__doc__



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ...._mars.core.entity import OutputType
from ...._mars.dataframe.utils import parse_index
from ...._mars.serialization.serializables import AnyField
from ...operand import DataOperand, DataOperandMixin


class HuggingfaceToDataframe(DataOperand, DataOperandMixin):
    types_mapper = AnyField("types_mapper")

    def __call__(self, dataset):
        params = dataset.params.copy()
        # dtypes is None trigger auto execution.
        if dataset.dtypes is not None:
            params["index_value"] = parse_index(pd.RangeIndex(-1))
            params["columns_value"] = parse_index(dataset.dtypes.index, store_data=True)
        return self.new_tileable([dataset], **params)

    @classmethod
    def tile(cls, op: "HuggingfaceToDataframe"):
        assert len(op.inputs) == 1
        inp = op.inputs[0]
        out = op.outputs[0]
        chunks = []
        for index, c in enumerate(inp.chunks):
            chunk_op = op.copy().reset_key()
            new_c = chunk_op.new_chunk(
                [c],
                index=(index, 0),
                shape=(np.nan, np.nan),
                columns_value=out.columns_value,
                index_value=parse_index(pd.RangeIndex(-1)),
                dtypes=out.dtypes,
            )
            chunks.append(new_c)
        return op.copy().new_tileable(
            op.inputs,
            chunks=chunks,
            nsplits=((np.nan,) * len(chunks), (np.nan,)),
            **out.params
        )

    @classmethod
    def execute(cls, ctx, op: "HuggingfaceToDataframe"):
        ds = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if op.types_mapper is None:
            df = ds.to_pandas()
        else:
            df = cls._to_pandas(ds, op.types_mapper)
        ctx[out.key] = df

    @classmethod
    def _to_pandas(cls, ds, types_mapper):
        from datasets.features.features import pandas_types_mapper
        from datasets.formatting import query_table

        def _types_mapper(_dtype):
            mapped = pandas_types_mapper(_dtype)
            if mapped is None:
                return types_mapper(_dtype)

        return query_table(
            table=ds._data,
            key=slice(0, len(ds)),
            indices=ds._indices if ds._indices is not None else None,
        ).to_pandas(types_mapper=_types_mapper)


def to_dataframe(dataset, types_mapper=None):
    return HuggingfaceToDataframe(
        output_types=[OutputType.dataframe], types_mapper=types_mapper
    )(dataset)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from numpy.compat import basestring

from ..datasource.array import tensor as astensor

__all__ = ["parse_einsum_input", "einsum_path"]

einsum_symbols = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
einsum_symbols_set = set(einsum_symbols)


def _flop_count(idx_contraction, inner, num_terms, size_dictionary):
    """
    Computes the number of FLOPS in the contraction.

    Parameters
    ----------
    idx_contraction : iterable
        The indices involved in the contraction
    inner : bool
        Does this contraction require an inner product?
    num_terms : int
        The number of terms in a contraction
    size_dictionary : dict
        The size of each of the indices in idx_contraction

    Returns
    -------
    flop_count : int
        The total number of FLOPS required for the contraction.

    Examples
    --------

    >>> _flop_count('abc', False, 1, {'a': 2, 'b':3, 'c':5})
    30

    >>> _flop_count('abc', True, 2, {'a': 2, 'b':3, 'c':5})
    60

    """

    overall_size = _compute_size_by_dict(idx_contraction, size_dictionary)
    op_factor = max(1, num_terms - 1)
    if inner:
        op_factor += 1

    return overall_size * op_factor


def _compute_size_by_dict(indices, idx_dict):
    """
    Computes the product of the elements in indices based on the dictionary
    idx_dict.

    Parameters
    ----------
    indices : iterable
        Indices to base the product on.
    idx_dict : dictionary
        Dictionary of index sizes

    Returns
    -------
    ret : int
        The resulting product.

    Examples
    --------
    >>> _compute_size_by_dict('abbc', {'a': 2, 'b':3, 'c':5})
    90

    """
    ret = 1
    for i in indices:
        ret *= idx_dict[i]
    return ret


def _find_contraction(positions, input_sets, output_set):
    """
    Finds the contraction for a given set of input and output sets.

    Parameters
    ----------
    positions : iterable
        Integer positions of terms used in the contraction.
    input_sets : list
        List of sets that represent the lhs side of the einsum subscript
    output_set : set
        Set that represents the rhs side of the overall einsum subscript

    Returns
    -------
    new_result : set
        The indices of the resulting contraction
    remaining : list
        List of sets that have not been contracted, the new set is appended to
        the end of this list
    idx_removed : set
        Indices removed from the entire contraction
    idx_contraction : set
        The indices used in the current contraction

    Examples
    --------

    # A simple dot product test case
    >>> pos = (0, 1)
    >>> isets = [set('ab'), set('bc')]
    >>> oset = set('ac')
    >>> _find_contraction(pos, isets, oset)
    ({'a', 'c'}, [{'a', 'c'}], {'b'}, {'a', 'b', 'c'})

    # A more complex case with additional terms in the contraction
    >>> pos = (0, 2)
    >>> isets = [set('abd'), set('ac'), set('bdc')]
    >>> oset = set('ac')
    >>> _find_contraction(pos, isets, oset)
    ({'a', 'c'}, [{'a', 'c'}, {'a', 'c'}], {'b', 'd'}, {'a', 'b', 'c', 'd'})
    """

    idx_contract = set()
    idx_remain = output_set.copy()
    remaining = []
    for ind, value in enumerate(input_sets):
        if ind in positions:
            idx_contract |= value
        else:
            remaining.append(value)
            idx_remain |= value

    new_result = idx_remain & idx_contract
    idx_removed = idx_contract - new_result
    remaining.append(new_result)

    return (new_result, remaining, idx_removed, idx_contract)


def _optimal_path(input_sets, output_set, idx_dict, memory_limit):
    """
    Computes all possible pair contractions, sieves the results based
    on ``memory_limit`` and returns the lowest cost path. This algorithm
    scales factorial with respect to the elements in the list ``input_sets``.

    Parameters
    ----------
    input_sets : list
        List of sets that represent the lhs side of the einsum subscript
    output_set : set
        Set that represents the rhs side of the overall einsum subscript
    idx_dict : dictionary
        Dictionary of index sizes
    memory_limit : int
        The maximum number of elements in a temporary array

    Returns
    -------
    path : list
        The optimal contraction order within the memory limit constraint.

    Examples
    --------
    >>> isets = [set('abd'), set('ac'), set('bdc')]
    >>> oset = set()
    >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}
    >>> _optimal_path(isets, oset, idx_sizes, 5000)
    [(0, 2), (0, 1)]
    """

    full_results = [(0, [], input_sets)]
    for iteration in range(len(input_sets) - 1):
        iter_results = []

        # Compute all unique pairs
        for curr in full_results:
            cost, positions, remaining = curr
            for con in itertools.combinations(range(len(input_sets) - iteration), 2):
                # Find the contraction
                cont = _find_contraction(con, remaining, output_set)
                new_result, new_input_sets, idx_removed, idx_contract = cont

                # Sieve the results based on memory_limit
                new_size = _compute_size_by_dict(new_result, idx_dict)
                if new_size > memory_limit:
                    continue

                # Build (total_cost, positions, indices_remaining)
                total_cost = cost + _flop_count(
                    idx_contract, idx_removed, len(con), idx_dict
                )
                new_pos = positions + [con]
                iter_results.append((total_cost, new_pos, new_input_sets))

        # Update combinatorial list, if we did not find anything return best
        # path + remaining contractions
        if iter_results:
            full_results = iter_results
        else:
            path = min(full_results, key=lambda x: x[0])[1]
            path += [tuple(range(len(input_sets) - iteration))]
            return path

    # If we have not found anything return single einsum contraction
    if len(full_results) == 0:
        return [tuple(range(len(input_sets)))]

    path = min(full_results, key=lambda x: x[0])[1]
    return path


def _parse_possible_contraction(
    positions, input_sets, output_set, idx_dict, memory_limit, path_cost, naive_cost
):
    """Compute the cost (removed size + flops) and resultant indices for
    performing the contraction specified by ``positions``.

    Parameters
    ----------
    positions : tuple of int
        The locations of the proposed tensors to contract.
    input_sets : list of sets
        The indices found on each tensors.
    output_set : set
        The output indices of the expression.
    idx_dict : dict
        Mapping of each index to its size.
    memory_limit : int
        The total allowed size for an intermediary tensor.
    path_cost : int
        The contraction cost so far.
    naive_cost : int
        The cost of the unoptimized expression.

    Returns
    -------
    cost : (int, int)
        A tuple containing the size of any indices removed, and the flop cost.
    positions : tuple of int
        The locations of the proposed tensors to contract.
    new_input_sets : list of sets
        The resulting new list of indices if this proposed contraction is performed.

    """

    # Find the contraction
    contract = _find_contraction(positions, input_sets, output_set)
    idx_result, new_input_sets, idx_removed, idx_contract = contract

    # Sieve the results based on memory_limit
    new_size = _compute_size_by_dict(idx_result, idx_dict)
    if new_size > memory_limit:
        return None

    # Build sort tuple
    old_sizes = (_compute_size_by_dict(input_sets[p], idx_dict) for p in positions)
    removed_size = sum(old_sizes) - new_size

    # NB: removed_size used to be just the size of any removed indices i.e.:
    #     helpers.compute_size_by_dict(idx_removed, idx_dict)
    cost = _flop_count(idx_contract, idx_removed, len(positions), idx_dict)
    sort = (-removed_size, cost)

    # Sieve based on total cost as well
    if (path_cost + cost) > naive_cost:
        return None

    # Add contraction to possible choices
    return [sort, positions, new_input_sets]


def _update_other_results(results, best):
    """Update the positions and provisional input_sets of ``results`` based on
    performing the contraction result ``best``. Remove any involving the tensors
    contracted.

    Parameters
    ----------
    results : list
        List of contraction results produced by ``_parse_possible_contraction``.
    best : list
        The best contraction of ``results`` i.e. the one that will be performed.

    Returns
    -------
    mod_results : list
        The list of modified results, updated with outcome of ``best`` contraction.
    """

    best_con = best[1]
    bx, by = best_con
    mod_results = []

    for cost, (x, y), con_sets in results:
        # Ignore results involving tensors just contracted
        if x in best_con or y in best_con:
            continue

        # Update the input_sets
        del con_sets[by - int(by > x) - int(by > y)]
        del con_sets[bx - int(bx > x) - int(bx > y)]
        con_sets.insert(-1, best[2][-1])

        # Update the position indices
        mod_con = x - int(x > bx) - int(x > by), y - int(y > bx) - int(y > by)
        mod_results.append((cost, mod_con, con_sets))

    return mod_results


def _greedy_path(input_sets, output_set, idx_dict, memory_limit):
    """
    Finds the path by contracting the best pair until the input list is
    exhausted. The best pair is found by minimizing the tuple
    ``(-prod(indices_removed), cost)``.  What this amounts to is prioritizing
    matrix multiplication or inner product operations, then Hadamard like
    operations, and finally outer operations. Outer products are limited by
    ``memory_limit``. This algorithm scales cubically with respect to the
    number of elements in the list ``input_sets``.

    Parameters
    ----------
    input_sets : list
        List of sets that represent the lhs side of the einsum subscript
    output_set : set
        Set that represents the rhs side of the overall einsum subscript
    idx_dict : dictionary
        Dictionary of index sizes
    memory_limit_limit : int
        The maximum number of elements in a temporary array

    Returns
    -------
    path : list
        The greedy contraction order within the memory limit constraint.

    Examples
    --------
    >>> isets = [set('abd'), set('ac'), set('bdc')]
    >>> oset = set()
    >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}
    >>> _greedy_path(isets, oset, idx_sizes, 5000)
    [(0, 2), (0, 1)]
    """

    # Handle trivial cases that leaked through
    if len(input_sets) == 1:
        return [(0,)]
    elif len(input_sets) == 2:
        return [(0, 1)]

    # Build up a naive cost
    contract = _find_contraction(range(len(input_sets)), input_sets, output_set)
    idx_result, new_input_sets, idx_removed, idx_contract = contract
    naive_cost = _flop_count(idx_contract, idx_removed, len(input_sets), idx_dict)

    # Initially iterate over all pairs
    comb_iter = itertools.combinations(range(len(input_sets)), 2)
    known_contractions = []

    path_cost = 0
    path = []

    for iteration in range(len(input_sets) - 1):
        # Iterate over all pairs on first step, only previously found pairs on subsequent steps
        for positions in comb_iter:
            # Always initially ignore outer products
            if input_sets[positions[0]].isdisjoint(input_sets[positions[1]]):
                continue

            result = _parse_possible_contraction(
                positions,
                input_sets,
                output_set,
                idx_dict,
                memory_limit,
                path_cost,
                naive_cost,
            )
            if result is not None:
                known_contractions.append(result)

        # If we do not have a inner contraction, rescan pairs including outer products
        if len(known_contractions) == 0:  # pragma: no cover
            # Then check the outer products
            for positions in itertools.combinations(range(len(input_sets)), 2):
                result = _parse_possible_contraction(
                    positions,
                    input_sets,
                    output_set,
                    idx_dict,
                    memory_limit,
                    path_cost,
                    naive_cost,
                )
                if result is not None:
                    known_contractions.append(result)

            # If we still did not find any remaining contractions, default back to einsum like behavior
            if len(known_contractions) == 0:
                path.append(tuple(range(len(input_sets))))
                break

        # Sort based on first index
        best = min(known_contractions, key=lambda x: x[0])

        # Now propagate as many unused contractions as possible to next iteration
        known_contractions = _update_other_results(known_contractions, best)

        # Next iteration only compute contractions with the new tensor
        # All other contractions have been accounted for
        input_sets = best[2]
        new_tensor_pos = len(input_sets) - 1
        comb_iter = ((i, new_tensor_pos) for i in range(new_tensor_pos))

        # Update path and total cost
        path.append(best[1])
        path_cost += best[0][1]

    return path


def _can_dot(inputs, result, idx_removed):
    """
    Checks if we can use BLAS (np.tensordot) call and its beneficial to do so.

    Parameters
    ----------
    inputs : list of str
        Specifies the subscripts for summation.
    result : str
        Resulting summation.
    idx_removed : set
        Indices that are removed in the summation


    Returns
    -------
    type : bool
        Returns true if BLAS should and can be used, else False

    Notes
    -----
    If the operations is BLAS level 1 or 2 and is not already aligned
    we default back to einsum as the memory movement to copy is more
    costly than the operation itself.


    Examples
    --------

    # Standard GEMM operation
    >>> _can_dot(['ij', 'jk'], 'ik', set('j'))
    True

    # Can use the standard BLAS, but requires odd data movement
    >>> _can_dot(['ijj', 'jk'], 'ik', set('j'))
    False

    # DDOT where the memory is not aligned
    >>> _can_dot(['ijk', 'ikj'], '', set('ijk'))
    False

    """

    # All `dot` calls remove indices
    if len(idx_removed) == 0:
        return False

    # BLAS can only handle two operands
    if len(inputs) != 2:
        return False

    input_left, input_right = inputs

    for c in set(input_left + input_right):
        # can't deal with repeated indices on same input or more than 2 total
        nl, nr = input_left.count(c), input_right.count(c)
        if (nl > 1) or (nr > 1) or (nl + nr > 2):
            return False

        # can't do implicit summation or dimension collapse e.g.
        #     "ab,bc->c" (implicitly sum over 'a')
        #     "ab,ca->ca" (take diagonal of 'a')
        if nl + nr - 1 == int(c in result):
            return False

    # Build a few temporaries
    set_left = set(input_left)
    set_right = set(input_right)
    keep_left = set_left - idx_removed
    keep_right = set_right - idx_removed
    rs = len(idx_removed)

    # At this point we are a DOT, GEMV, or GEMM operation

    # Handle inner products

    # DDOT with aligned data
    if input_left == input_right:
        return True

    # DDOT without aligned data (better to use einsum)
    if set_left == set_right:
        return False

    # Handle the 4 possible (aligned) GEMV or GEMM cases

    # GEMM or GEMV no transpose
    if input_left[-rs:] == input_right[:rs]:
        return True

    # GEMM or GEMV transpose both
    if input_left[:rs] == input_right[-rs:]:
        return True

    # GEMM or GEMV transpose right
    if input_left[-rs:] == input_right[-rs:]:
        return True

    # GEMM or GEMV transpose left
    if input_left[:rs] == input_right[:rs]:
        return True

    # Einsum is faster than GEMV if we have to copy data
    if not keep_left or not keep_right:
        return False

    # We are a matrix-matrix product, but we need to copy data
    return True


def parse_einsum_input(operands):
    """
    A reproduction of einsum c side einsum parsing in python.

    Returns
    -------
    input_strings : str
        Parsed input strings
    output_string : str
        Parsed output string
    operands : list of array_like
        The operands to use in the numpy contraction

    Examples
    --------
    The operand list is simplified to reduce printing:

    >>> import mars.tensor as mt
    >>> mt.random.seed(123)
    >>> a = mt.random.rand(4, 4)
    >>> b = mt.random.rand(4, 4, 4)
    >>> parse_einsum_input(('...a,...a->...', a, b))
    ('za,xza', 'xz', [a, b]) # may vary

    >>> parse_einsum_input((a, [Ellipsis, 0], b, [Ellipsis, 0]))
    ('za,xza', 'xz', [a, b]) # may vary
    """

    if len(operands) == 0:
        raise ValueError("No input operands")

    if isinstance(operands[0], basestring):
        subscripts = operands[0].replace(" ", "")
        operands = [astensor(v) for v in operands[1:]]

        # Ensure all characters are valid
        for s in subscripts:
            if s in ".,->":
                continue
            if s not in einsum_symbols:
                raise ValueError(f"Character {s} is not a valid symbol.")

    else:  # pragma: no cover
        tmp_operands = list(operands)
        operand_list = []
        subscript_list = []
        for p in range(len(operands) // 2):
            operand_list.append(tmp_operands.pop(0))
            subscript_list.append(tmp_operands.pop(0))

        output_list = tmp_operands[-1] if len(tmp_operands) else None
        operands = [astensor(v) for v in operand_list]
        subscripts = ""
        last = len(subscript_list) - 1
        for num, sub in enumerate(subscript_list):
            for s in sub:
                if s is Ellipsis:
                    subscripts += "..."
                elif isinstance(s, int):
                    subscripts += einsum_symbols[s]
                else:
                    raise TypeError(
                        "For this input type lists must contain "
                        "either int or Ellipsis"
                    )
            if num != last:
                subscripts += ","

        if output_list is not None:
            subscripts += "->"
            for s in output_list:
                if s is Ellipsis:
                    subscripts += "..."
                elif isinstance(s, int):
                    subscripts += einsum_symbols[s]
                else:
                    raise TypeError(
                        "For this input type lists must contain "
                        "either int or Ellipsis"
                    )
    # Check for proper "->"
    if ("-" in subscripts) or (">" in subscripts):
        invalid = (subscripts.count("-") > 1) or (subscripts.count(">") > 1)
        if invalid or (subscripts.count("->") != 1):
            raise ValueError("Subscripts can only contain one '->'.")

    # Parse ellipses
    if "." in subscripts:
        used = subscripts.replace(".", "").replace(",", "").replace("->", "")
        unused = list(einsum_symbols_set - set(used))
        ellipse_inds = "".join(unused)
        longest = 0

        if "->" in subscripts:
            input_tmp, output_sub = subscripts.split("->")
            split_subscripts = input_tmp.split(",")
            out_sub = True
        else:
            split_subscripts = subscripts.split(",")
            out_sub = False

        for num, sub in enumerate(split_subscripts):
            if "." in sub:
                if (sub.count(".") != 3) or (sub.count("...") != 1):
                    raise ValueError("Invalid Ellipses.")

                # Take into account numerical values
                if operands[num].shape == ():
                    ellipse_count = 0
                else:
                    ellipse_count = max(operands[num].ndim, 1)
                    ellipse_count -= len(sub) - 3

                if ellipse_count > longest:
                    longest = ellipse_count

                if ellipse_count < 0:
                    raise ValueError("Ellipses lengths do not match.")
                elif ellipse_count == 0:
                    split_subscripts[num] = sub.replace("...", "")
                else:
                    rep_inds = ellipse_inds[-ellipse_count:]
                    split_subscripts[num] = sub.replace("...", rep_inds)

        subscripts = ",".join(split_subscripts)
        if longest == 0:
            out_ellipse = ""
        else:
            out_ellipse = ellipse_inds[-longest:]

        if out_sub:
            subscripts += "->" + output_sub.replace("...", out_ellipse)
        else:
            # Special care for outputless ellipses
            output_subscript = ""
            tmp_subscripts = subscripts.replace(",", "")
            for s in sorted(set(tmp_subscripts)):
                if s not in (einsum_symbols):
                    raise ValueError(f"Character {s} is not a valid symbol.")
                if tmp_subscripts.count(s) == 1:
                    output_subscript += s
            normal_inds = "".join(sorted(set(output_subscript) - set(out_ellipse)))

            subscripts += "->" + out_ellipse + normal_inds

    # Build output string if does not exist
    if "->" in subscripts:
        input_subscripts, output_subscript = subscripts.split("->")
    else:
        input_subscripts = subscripts
        # Build output subscripts
        tmp_subscripts = subscripts.replace(",", "")
        output_subscript = ""
        for s in sorted(set(tmp_subscripts)):
            if s not in einsum_symbols:
                raise ValueError(f"Character {s} is not a valid symbol.")
            if tmp_subscripts.count(s) == 1:
                output_subscript += s

    # Make sure output subscripts are in the input
    for char in output_subscript:
        if char not in input_subscripts:
            raise ValueError(f"Output character {char} did not appear in the input")

    # Make sure number operands is equivalent to the number of terms
    if len(input_subscripts.split(",")) != len(operands):
        raise ValueError(
            "Number of einsum subscripts must be equal to the number of operands."
        )

    return (input_subscripts, output_subscript, operands)


def _einsum_path_dispatcher(*operands, **kwargs):
    # NOTE: technically, we should only dispatch on array-like arguments, not
    # subscripts (given as strings). But separating operands into
    # arrays/subscripts is a little tricky/slow (given einsum's two supported
    # signatures), so as a practical shortcut we dispatch on everything.
    # Strings will be ignored for dispatching since they don't define
    # __array_function__.
    return operands


def einsum_path(*operands, **kwargs):
    """
    einsum_path(subscripts, *operands, optimize='greedy')

    Evaluates the lowest cost contraction order for an einsum expression by
    considering the creation of intermediate arrays.

    Parameters
    ----------
    subscripts : str
        Specifies the subscripts for summation.
    *operands : list of array_like
        These are the arrays for the operation.
    optimize : {bool, list, tuple, 'greedy', 'optimal'}
        Choose the type of path. If a tuple is provided, the second argument is
        assumed to be the maximum intermediate size created. If only a single
        argument is provided the largest input or output array size is used
        as a maximum intermediate size.

        * if a list is given that starts with ``einsum_path``, uses this as the
          contraction path
        * if False no optimization is taken
        * if True defaults to the 'greedy' algorithm
        * 'optimal' An algorithm that combinatorially explores all possible
          ways of contracting the listed tensors and choosest the least costly
          path. Scales exponentially with the number of terms in the
          contraction.
        * 'greedy' An algorithm that chooses the best pair contraction
          at each step. Effectively, this algorithm searches the largest inner,
          Hadamard, and then outer products at each step. Scales cubically with
          the number of terms in the contraction. Equivalent to the 'optimal'
          path for most contractions.

        Default is 'greedy'.

    Returns
    -------
    path : list of tuples
        A list representation of the einsum path.
    string_repr : str
        A printable representation of the einsum path.

    Notes
    -----
    The resulting path indicates which terms of the input contraction should be
    contracted first, the result of this contraction is then appended to the
    end of the contraction list. This list can then be iterated over until all
    intermediate contractions are complete.

    See Also
    --------
    einsum, linalg.multi_dot

    Examples
    --------

    We can begin with a chain dot example. In this case, it is optimal to
    contract the ``b`` and ``c`` tensors first as represented by the first
    element of the path ``(1, 2)``. The resulting tensor is added to the end
    of the contraction and the remaining contraction ``(0, 1)`` is then
    completed.

    >>> np.random.seed(123)
    >>> a = np.random.rand(2, 2)
    >>> b = np.random.rand(2, 5)
    >>> c = np.random.rand(5, 2)
    >>> path_info = np.einsum_path('ij,jk,kl->il', a, b, c, optimize='greedy')
    >>> print(path_info[0])
    ['einsum_path', (1, 2), (0, 1)]
    >>> print(path_info[1])
      Complete contraction:  ij,jk,kl->il # may vary
             Naive scaling:  4
         Optimized scaling:  3
          Naive FLOP count:  1.600e+02
      Optimized FLOP count:  5.600e+01
       Theoretical speedup:  2.857
      Largest intermediate:  4.000e+00 elements
    -------------------------------------------------------------------------
    scaling                  current                                remaining
    -------------------------------------------------------------------------
       3                   kl,jk->jl                                ij,jl->il
       3                   jl,ij->il                                   il->il


    A more complex index transformation example.

    >>> I = np.random.rand(10, 10, 10, 10)
    >>> C = np.random.rand(10, 10)
    >>> path_info = np.einsum_path('ea,fb,abcd,gc,hd->efgh', C, C, I, C, C,
    ...                            optimize='greedy')

    >>> print(path_info[0])
    ['einsum_path', (0, 2), (0, 3), (0, 2), (0, 1)]
    >>> print(path_info[1])
      Complete contraction:  ea,fb,abcd,gc,hd->efgh # may vary
             Naive scaling:  8
         Optimized scaling:  5
          Naive FLOP count:  8.000e+08
      Optimized FLOP count:  8.000e+05
       Theoretical speedup:  1000.000
      Largest intermediate:  1.000e+04 elements
    --------------------------------------------------------------------------
    scaling                  current                                remaining
    --------------------------------------------------------------------------
       5               abcd,ea->bcde                      fb,gc,hd,bcde->efgh
       5               bcde,fb->cdef                         gc,hd,cdef->efgh
       5               cdef,gc->defg                            hd,defg->efgh
       5               defg,hd->efgh                               efgh->efgh
    """

    # Make sure all keywords are valid
    valid_contract_kwargs = ["optimize", "einsum_call"]
    unknown_kwargs = [k for (k, v) in kwargs.items() if k not in valid_contract_kwargs]
    if len(unknown_kwargs):
        raise TypeError(f"Did not understand the following kwargs: {unknown_kwargs!r}")

    # Figure out what the path really is
    path_type = kwargs.pop("optimize", True)
    if path_type is True:
        path_type = "greedy"
    if path_type is None:
        path_type = False

    memory_limit = None

    # No optimization or a named path algorithm
    if (path_type is False) or isinstance(path_type, basestring):
        pass

    # Given an explicit path
    elif len(path_type) and (path_type[0] == "einsum_path"):  # pragma: no cover
        pass

    # Path tuple with memory limit
    elif (
        (len(path_type) == 2)
        and isinstance(path_type[0], basestring)
        and isinstance(path_type[1], (int, float))
    ):  # pragma: no cover
        memory_limit = int(path_type[1])
        path_type = path_type[0]

    else:  # pragma: no cover
        raise TypeError(f"Did not understand the path: {path_type}")

    # Hidden option, only einsum should call this
    einsum_call_arg = kwargs.pop("einsum_call", False)

    # Python side parsing
    input_subscripts, output_subscript, operands = parse_einsum_input(operands)

    # Build a few useful list and sets
    input_list = input_subscripts.split(",")
    input_sets = [set(x) for x in input_list]
    output_set = set(output_subscript)
    indices = set(input_subscripts.replace(",", ""))

    # Get length of each unique dimension and ensure all dimensions are correct
    dimension_dict = {}
    broadcast_indices = [[] for x in range(len(input_list))]
    for tnum, term in enumerate(input_list):
        sh = operands[tnum].shape
        if len(sh) != len(term):
            raise ValueError(
                "Einstein sum subscript %s does not contain the "
                "correct number of indices for operand %d."
                % (input_subscripts[tnum], tnum)
            )
        for cnum, char in enumerate(term):
            dim = sh[cnum]

            # Build out broadcast indices
            if dim == 1:
                broadcast_indices[tnum].append(char)

            if char in dimension_dict.keys():
                # For broadcasting cases we always want the largest dim size
                if dimension_dict[char] == 1:
                    dimension_dict[char] = dim
                elif dim not in (1, dimension_dict[char]):
                    raise ValueError(
                        "Size of label '%s' for operand %d (%d) "
                        "does not match previous terms (%d)."
                        % (char, tnum, dimension_dict[char], dim)
                    )
            else:
                dimension_dict[char] = dim

    # Convert broadcast inds to sets
    broadcast_indices = [set(x) for x in broadcast_indices]

    # Compute size of each input array plus the output array
    size_list = [
        _compute_size_by_dict(term, dimension_dict)
        for term in input_list + [output_subscript]
    ]
    max_size = max(size_list)

    if memory_limit is None:
        memory_arg = max_size
    else:
        memory_arg = memory_limit

    # Compute naive cost
    # This isn't quite right, need to look into exactly how einsum does this
    inner_product = (sum(len(x) for x in input_sets) - len(indices)) > 0
    naive_cost = _flop_count(indices, inner_product, len(input_list), dimension_dict)

    # Compute the path
    if (path_type is False) or (len(input_list) in [1, 2]) or (indices == output_set):
        # Nothing to be optimized, leave it to einsum
        path = [tuple(range(len(input_list)))]
    elif path_type == "greedy":
        path = _greedy_path(input_sets, output_set, dimension_dict, memory_arg)
    elif path_type == "optimal":
        path = _optimal_path(input_sets, output_set, dimension_dict, memory_arg)
    elif path_type[0] == "einsum_path":  # pragma: no cover
        path = path_type[1:]
    else:  # pragma: no cover
        raise KeyError("Path name %s not found", path_type)

    cost_list, scale_list, size_list, contraction_list = [], [], [], []

    # Build contraction tuple (positions, gemm, einsum_str, remaining)
    for cnum, contract_inds in enumerate(path):
        # Make sure we remove inds from right to left
        contract_inds = tuple(sorted(list(contract_inds), reverse=True))

        contract = _find_contraction(contract_inds, input_sets, output_set)
        out_inds, input_sets, idx_removed, idx_contract = contract

        cost = _flop_count(
            idx_contract, idx_removed, len(contract_inds), dimension_dict
        )
        cost_list.append(cost)
        scale_list.append(len(idx_contract))
        size_list.append(_compute_size_by_dict(out_inds, dimension_dict))

        bcast = set()
        tmp_inputs = []
        for x in contract_inds:
            tmp_inputs.append(input_list.pop(x))
            bcast |= broadcast_indices.pop(x)

        new_bcast_inds = bcast - idx_removed

        # If we're broadcasting, nix blas
        if not len(idx_removed & bcast):
            do_blas = _can_dot(tmp_inputs, out_inds, idx_removed)
        else:
            do_blas = False

        # Last contraction
        if (cnum - len(path)) == -1:
            idx_result = output_subscript
        else:
            sort_result = [(dimension_dict[ind], ind) for ind in out_inds]
            idx_result = "".join([x[1] for x in sorted(sort_result)])

        input_list.append(idx_result)
        broadcast_indices.append(new_bcast_inds)
        einsum_str = ",".join(tmp_inputs) + "->" + idx_result

        contraction = (contract_inds, idx_removed, einsum_str, input_list[:], do_blas)
        contraction_list.append(contraction)

    opt_cost = sum(cost_list) + 1

    if einsum_call_arg:
        return (operands, contraction_list)

    # Return the path along with a nice string representation
    overall_contraction = input_subscripts + "->" + output_subscript
    header = ("scaling", "current", "remaining")

    speedup = naive_cost / opt_cost
    max_i = max(size_list)

    path_print = "  Complete contraction:  %s\n" % overall_contraction  # noqa: E221
    path_print += "         Naive scaling:  %d\n" % len(indices)
    path_print += "     Optimized scaling:  %d\n" % max(scale_list)
    path_print += "      Naive FLOP count:  %.3e\n" % naive_cost
    path_print += "  Optimized FLOP count:  %.3e\n" % opt_cost
    path_print += "   Theoretical speedup:  %3.3f\n" % speedup
    path_print += "  Largest intermediate:  %.3e elements\n" % max_i
    path_print += "-" * 74 + "\n"
    path_print += "%6s %24s %40s\n" % header
    path_print += "-" * 74

    for n, contraction in enumerate(contraction_list):
        inds, idx_rm, einsum_str, remaining, blas = contraction
        remaining_str = ",".join(remaining) + "->" + output_subscript
        path_run = (scale_list[n], einsum_str, remaining_str)
        path_print += "\n%4d    %24s %40s" % path_run

    path = ["einsum_path"] + path
    return path, path_print



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
import sys

from xoscar.aio.file import AioFileObject
from xoscar.aio.lru import alru_cache
from xoscar.aio.parallelism import AioEvent

from .file import AioFilesystem
from .isolation import Isolation, get_isolation, new_isolation, stop_isolation

if sys.version_info[:2] < (3, 9):
    from ._threads import to_thread

    asyncio.to_thread = to_thread



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.neighbors as sk_neighbors

from ...._mars.learn import neighbors as mars_neighbors
from ...._mars.learn.neighbors import NearestNeighbors as MarsNearestNeighbors
from ....core.utils.docstring import attach_module_callable_docstring
from ...utils import SKLearnBase, _collect_module_callables, _install_cls_members


class NearestNeighbors(SKLearnBase):
    _marscls = MarsNearestNeighbors


SKLEARN_NEIGHBORS_CLS_MAP = {
    NearestNeighbors: MarsNearestNeighbors,
}

MARS_SKLEARN_NEIGHBORS_CALLABLES = _collect_module_callables(
    mars_neighbors, sk_neighbors, skip_members=["register_op"]
)
_install_cls_members(
    SKLEARN_NEIGHBORS_CLS_MAP, MARS_SKLEARN_NEIGHBORS_CALLABLES, sk_neighbors
)
attach_module_callable_docstring(
    NearestNeighbors, sk_neighbors, sk_neighbors.NearestNeighbors
)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import warnings
from functools import lru_cache
from types import ModuleType
from typing import Any, Callable, Dict, Type

import numpy as np
import scipy

from ..._mars.core import Entity as MarsEntity
from ...core import DataType
from ...core.adapter import (
    ClsMethodWrapper,
    MarsOutputType,
    get_cls_members,
    wrap_mars_callable,
)
from ...core.utils.fallback import wrap_fallback_module_method

_NO_ANNOTATION_FUNCS: Dict[Callable, MarsOutputType] = {
    np.busday_count: MarsOutputType.tensor,
    np.busday_offset: MarsOutputType.object,
    np.is_busday: MarsOutputType.object,
    np.isneginf: MarsOutputType.object,
    np.isposinf: MarsOutputType.object,
    np.einsum_path: MarsOutputType.object,
    np.kron: MarsOutputType.tensor,
    np.outer: MarsOutputType.tensor,
    np.trace: MarsOutputType.tensor,
    np.linalg.cond: MarsOutputType.tensor,
    np.linalg.det: MarsOutputType.tensor,
    np.linalg.eig: MarsOutputType.object,
    np.linalg.eigh: MarsOutputType.object,
    np.linalg.eigvals: MarsOutputType.tensor,
    np.linalg.eigvalsh: MarsOutputType.tensor,
    np.linalg.tensorsolve: MarsOutputType.tensor,
    np.linalg.multi_dot: MarsOutputType.tensor,
    np.linalg.matrix_power: MarsOutputType.tensor,
    np.linalg.matrix_rank: MarsOutputType.tensor,
    np.linalg.lstsq: MarsOutputType.object,
    np.linalg.slogdet: MarsOutputType.object,
    np.linalg.pinv: MarsOutputType.tensor,
    np.linalg.tensorinv: MarsOutputType.tensor,
    np.random.default_rng: MarsOutputType.object,
    np.random.Generator: MarsOutputType.object,
    np.random.PCG64: MarsOutputType.object,
    np.random.MT19937: MarsOutputType.object,
    np.random.Generator: MarsOutputType.object,
}


class NumpyClsMethodWrapper(ClsMethodWrapper):
    def _generate_fallback_data(self, mars_entity: MarsEntity):
        return mars_entity.to_numpy()

    def _generate_warning_msg(self, entity: MarsEntity, func_name: str):
        return f"{type(entity).__name__}.{func_name} will fallback to Numpy"

    def _get_output_type(self, func: Callable) -> MarsOutputType:
        try:  # pragma: no cover
            return_annotation = inspect.signature(func).return_annotation
            if return_annotation is inspect.Signature.empty:
                # mostly for python3.7 whose return_annotation is always empty
                return _NO_ANNOTATION_FUNCS.get(func, MarsOutputType.object)
            all_types = [t.strip() for t in return_annotation.split("|")]

            return (
                MarsOutputType.tensor
                if "ndarray" in all_types
                else MarsOutputType.object
            )
        except (
            ValueError
        ):  # some np methods return objects and inspect.signature throws a ValueError
            return _NO_ANNOTATION_FUNCS.get(func, MarsOutputType.object)

    def _get_docstring_src_module(self):
        return np


def _collect_numpy_cls_members(np_cls: Type, data_type: DataType):
    members = get_cls_members(data_type)
    for name, np_cls_member in inspect.getmembers(np_cls):
        if name not in members and not name.startswith("_"):
            numpy_cls_method_wrapper = NumpyClsMethodWrapper(
                library_cls=np_cls, func_name=name, fallback_warning=True
            )
            members[name] = numpy_cls_method_wrapper.get_wrapped()


def _collect_numpy_ndarray_members():
    _collect_numpy_cls_members(np.ndarray, DataType.tensor)


@lru_cache(maxsize=1)
def collect_numpy_module_members(np_mod: ModuleType) -> Dict[str, Any]:
    from ..mars_adapters.core import MARS_TENSOR_CALLABLES

    module_methods: Dict[str, Any] = dict()
    with warnings.catch_warnings():
        for name, mod_member in inspect.getmembers(np_mod):
            if (
                name not in MARS_TENSOR_CALLABLES
                and callable(mod_member)
                and not name.startswith("_")
            ):
                # avoid inconsistency: np.ramdom.__name__ is 'numpy.random' while np.ndarray.__name__ is 'ndarray'
                np_mod_str = (
                    np_mod.__name__
                    if "numpy" in np_mod.__name__
                    else "numpy." + np_mod.__name__
                )
                warning_str = f"xorbits.{np_mod_str}.{name} will fallback to NumPy"
                numpy_cls_method_wrapper = NumpyClsMethodWrapper()
                output_type = numpy_cls_method_wrapper._get_output_type(
                    func=getattr(np_mod, name)
                )

                module_methods[name] = wrap_mars_callable(
                    wrap_fallback_module_method(np_mod, name, output_type, warning_str),
                    attach_docstring=True,
                    is_cls_member=False,
                    docstring_src_module=np_mod,
                    docstring_src=getattr(np_mod, name, None),
                    fallback_warning=True,
                )

    return module_methods


NUMPY_MEMBERS = collect_numpy_module_members(np)
NUMPY_LINALG_MEMBERS = collect_numpy_module_members(np.linalg)
NUMPY_FFT_MEMBERS = collect_numpy_module_members(np.fft)
NUMPY_RANDOM_MEMBERS = collect_numpy_module_members(np.random)
NUMPY_SPECIAL_MEMBERS = collect_numpy_module_members(scipy.special)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import numpy as np

from ...._mars.core.entity import OutputType
from ...._mars.serialization.serializables import AnyField, DictField
from ...operand import DataOperand, DataOperandMixin


class HuggingfaceMap(DataOperand, DataOperandMixin):
    func = AnyField("func")
    # The arguments are passing to huggingface dataset map:
    # https://huggingface.co/docs/datasets/v2.13.1/en/package_reference/main_classes#datasets.Dataset.map
    hf_kwargs = DictField("hf_kwargs")

    def __call__(self, dataset):
        # serialize in advance to reduce overhead
        self.func = cloudpickle.dumps(self.func)
        return self.new_tileable([dataset], **dataset.params)

    @classmethod
    def tile(cls, op: "HuggingfaceMap"):
        assert len(op.inputs) == 1
        inp = op.inputs[0]
        out = op.outputs[0]
        out_chunks = []
        for chunk in inp.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([chunk], index=chunk.index)
            out_chunks.append(out_chunk)
        return op.copy().new_tileable(
            op.inputs,
            chunks=out_chunks,
            nsplits=((np.nan,) * len(out_chunks), (np.nan,)),
            **out.params,
        )

    @classmethod
    def execute(cls, ctx, op: "HuggingfaceMap"):
        func = cloudpickle.loads(op.func)
        inp = ctx[op.inputs[0].key]
        out = op.outputs[0]
        ctx[out.key] = inp.map(func, **op.hf_kwargs)


def map(dataset, fn, **hf_kwargs):
    op = HuggingfaceMap(
        output_types=[OutputType.huggingface_dataset], func=fn, hf_kwargs=hf_kwargs
    )
    return op(dataset)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils import check_consistent_length
from ..utils import make_import_error_func
from ._predict import predict_base
from ._train import train
from .core import LGBMModelType, LGBMScikitLearnBase

try:
    import lightgbm
except ImportError:
    lightgbm = None


if not lightgbm:
    LGBMRegressor = make_import_error_func("lightgbm")
else:

    class LGBMRegressor(LGBMScikitLearnBase, lightgbm.LGBMRegressor):
        def fit(
            self,
            X,
            y,
            sample_weight=None,
            init_score=None,
            eval_set=None,
            eval_sample_weight=None,
            eval_init_score=None,
            session=None,
            run_kwargs=None,
            **kwargs
        ):
            check_consistent_length(X, y, session=session, run_kwargs=run_kwargs)
            params = self.get_params(True)
            model = train(
                params,
                self._wrap_train_tuple(X, y, sample_weight, init_score),
                eval_sets=self._wrap_eval_tuples(
                    eval_set, eval_sample_weight, eval_init_score
                ),
                model_type=LGBMModelType.REGRESSOR,
                session=session,
                run_kwargs=run_kwargs,
                **kwargs
            )

            self.set_params(**model.get_params())
            self._copy_extra_params(model, self)
            return self

        def predict(self, X, **kw):
            session = kw.pop("session", None)
            run_kwargs = kw.pop("run_kwargs", None)
            X = self._convert_tileable(X)
            return predict_base(self, X, session=session, run_kwargs=run_kwargs, **kw)

        def to_local(self):
            model = lightgbm.LGBMRegressor(**self.get_params())
            self._copy_extra_params(self, model)
            return model



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .arange import TensorArange, arange
from .array import (
    ArrayDataSource,
    CSRMatrixDataSource,
    array,
    asarray,
    ascontiguousarray,
    asfortranarray,
    tensor,
)
from .diag import TensorDiag, diag
from .diagflat import diagflat
from .empty import TensorEmpty, TensorEmptyLike, empty, empty_like
from .eye import TensorEye, eye
from .from_dataframe import TensorFromDataFrame, from_dataframe, from_series
from .from_dense import DenseToSparse, fromdense
from .from_hdf5 import TensorHDF5DataSource, fromhdf5
from .from_sparse import SparseToDense, fromsparse
from .from_tiledb import TensorTileDBDataSource, fromtiledb
from .from_vineyard import TensorFromVineyard, TensorFromVineyardChunk, fromvineyard
from .from_zarr import TensorFromZarr, fromzarr
from .full import TensorFull, TensorFullLike, full, full_like
from .identity import identity
from .indices import TensorIndices, indices
from .linspace import TensorLinspace, linspace
from .meshgrid import meshgrid
from .ones import TensorOnes, TensorOnesLike, ones, ones_like
from .scalar import Scalar, scalar
from .tri import TensorTril, TensorTriu, tri, tril, triu
from .zeros import TensorZeros, TensorZerosLike, zeros, zeros_like



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import functools
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Set, Tuple, Type, Union

from ...core import ChunkGraph, ChunkType, OperandType
from ...utils import build_fuse_chunk

REDUCTION = object()


@dataclasses.dataclass
class _Fuse:
    graph: ChunkGraph
    heads: List[ChunkType]
    tails: List[ChunkType]


class RuntimeOptimizer(ABC):
    engine = None

    def __init__(self, graph: ChunkGraph):
        self._graph = graph

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        """
        Check this optimizer is available.

        Returns
        -------
        is_available : bool
            Available.
        """

    @abstractmethod
    def optimize(self) -> Tuple[List[_Fuse], List[ChunkType]]:
        """
        Optimize chunk graph.
        """

    def _fuse_nodes(
        self, fuses: List[List[ChunkType]], fuse_cls: OperandType
    ) -> Tuple[List[List[ChunkType]], List[ChunkType]]:
        graph = self._graph
        fused_nodes = []

        for fuse in fuses:
            head_node = fuse[0]
            tail_node = fuse[-1]

            fused_chunk = build_fuse_chunk(
                fuse, fuse_cls, op_kw={"dtype": tail_node.dtype}
            ).data
            graph.add_node(fused_chunk)
            for node in graph.iter_successors(tail_node):
                graph.add_edge(fused_chunk, node)
            for node in graph.iter_predecessors(head_node):
                graph.add_edge(node, fused_chunk)
            for node in fuse:
                graph.remove_node(node)
            fused_nodes.append(fused_chunk)

            try:
                # check tail node if it's in results
                i = graph.results.index(tail_node)
                graph.results[i] = fused_chunk
            except ValueError:
                pass

        return fuses, fused_nodes


class GraphTraversalOptimizer(RuntimeOptimizer):
    @abstractmethod
    def _can_fuse(self, node: ChunkType) -> Union[bool, object]:
        """
        Check if a node can be fused

        Returns
        -------
        bool or object:
            If the node can be fused and is a reduction operation, return
            REDUCTION. Otherwise, if the node can be fused and is not a
            reduction operation, return True. For other nodes, return False
        """

    def _collect_fuse(
        self,
        graph: ChunkGraph,
        node: ChunkType,
        graph_results: Set[ChunkType],
        cached_can_fuse: Callable,
    ) -> _Fuse:
        fuse_graph = ChunkGraph()
        fuse_graph.add_node(node)
        fuse_heads: ChunkType = []
        fuse_tails: ChunkType = []
        tail_reduction_node: ChunkType = None

        stack: List[ChunkType] = [node]
        # Do a full search of sub graph even the fuse tails > 1
        while len(stack) != 0:
            node = stack.pop()
            is_head = graph.count_predecessors(node) == 0
            for n in graph.iter_predecessors(node):
                can_fuse = cached_can_fuse(n)
                if can_fuse is False or can_fuse is REDUCTION:
                    is_head = True
                elif not fuse_graph.contains(n):
                    stack.append(n)
                    fuse_graph.add_node(n)
                else:
                    fuse_graph.add_edge(n, node)
            if is_head:
                fuse_heads.append(node)
            # Skip the successors of tail reduction node.
            if node is tail_reduction_node:
                continue
            is_tail = graph.count_successors(node) == 0 or node in graph_results
            for n in graph.iter_successors(node):
                can_fuse = cached_can_fuse(n)
                if can_fuse is False:
                    is_tail = True
                elif can_fuse is REDUCTION:
                    if tail_reduction_node is None:
                        tail_reduction_node = n
                        fuse_tails.append(n)
                        stack.append(n)
                        fuse_graph.add_node(n)
                    elif n is tail_reduction_node:
                        fuse_graph.add_edge(node, n)
                    else:
                        is_tail = True
                elif not fuse_graph.contains(n):
                    stack.append(n)
                    fuse_graph.add_node(n)
                else:
                    fuse_graph.add_edge(node, n)
            if is_tail:
                fuse_tails.append(node)

        return _Fuse(fuse_graph, fuse_heads, fuse_tails)

    def _graph_traverse(self, logger, logger_str: str) -> List[_Fuse]:
        fuses: List[_Fuse] = []
        explored: Set[ChunkType] = set()
        cached_can_fuse = functools.lru_cache(maxsize=None)(self._can_fuse)

        graph = self._graph
        graph_results = set(graph.results)
        for node in graph.topological_iter():
            if node.op.gpu or node.op.sparse:
                # break
                return [], []
            if node in explored or node in graph_results:
                continue
            can_fuse = cached_can_fuse(node)
            if can_fuse is True:
                fuse = self._collect_fuse(graph, node, graph_results, cached_can_fuse)
                if len(fuse.graph) > 1:
                    explored.update(fuse.graph)
                    if len(fuse.tails) == 1:
                        fuses.append(fuse)
                    else:
                        logger.info(logger_str)
        return fuses

    def _fuse_nodes(
        self, fuses: List[_Fuse], fuse_cls
    ) -> Tuple[List[_Fuse], List[ChunkType]]:
        graph = self._graph
        fused_nodes: List[ChunkType] = []

        for fuse in fuses:
            fuse_graph = fuse.graph
            tail_nodes = fuse.tails
            head_nodes = fuse.heads
            inputs = [
                inp for n in head_nodes for inp in n.inputs if inp not in fuse_graph
            ]

            tail_chunk = tail_nodes[0]
            tail_chunk_op = tail_chunk.op
            fuse_op = fuse_cls(
                sparse=tail_chunk_op.sparse,
                gpu=tail_chunk_op.gpu,
                _key=tail_chunk_op.key,
                fuse_graph=fuse_graph,
                dtype=tail_chunk.dtype,
            )
            fused_chunk = fuse_op.new_chunk(
                inputs,
                kws=[tail_chunk.params],
                _key=tail_chunk.key,
                _chunk=tail_chunk,
            ).data

            graph.add_node(fused_chunk)
            for node in graph.iter_successors(tail_chunk):
                graph.add_edge(fused_chunk, node)
            for head_chunk in head_nodes:
                for node in graph.iter_predecessors(head_chunk):
                    if not fuse_graph.contains(node):
                        graph.add_edge(node, fused_chunk)
            for node in fuse_graph:
                graph.remove_node(node)
            fused_nodes.append(fused_chunk)

            try:
                # check tail node if it's in results
                i = graph.results.index(tail_chunk)
                graph.results[i] = fused_chunk
            except ValueError:
                pass

        return fuses, fused_nodes


_engine_to_optimizers: Dict[str, Type[RuntimeOptimizer]] = dict()


def register_optimizer(optimizer_cls: Type[RuntimeOptimizer]):
    _engine_to_optimizers[optimizer_cls.engine] = optimizer_cls
    return optimizer_cls


def optimize(graph: ChunkGraph, engines: List[str] = None) -> ChunkGraph:
    if engines is None:
        engines = ["jax", "numexpr", "cupy"]

    for engine in engines:
        optimizer_cls = _engine_to_optimizers[engine]
        optimizer = optimizer_cls(graph)
        if not optimizer.is_available():
            continue
        optimizer.optimize()

    return graph



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.decomposition as sk_decomposition

from ...._mars.learn import decomposition as mars_decomposition
from ...._mars.learn.decomposition import PCA as MarsPCA
from ...._mars.learn.decomposition import TruncatedSVD as MarsTruncatedSVD
from ....core.utils.docstring import attach_module_callable_docstring
from ...utils import SKLearnBase, _collect_module_callables, _install_cls_members


class PCA(SKLearnBase):
    _marscls = MarsPCA


class TruncatedSVD(SKLearnBase):
    _marscls = MarsTruncatedSVD


SKLEARN_DECOMP_CLS_MAP = {PCA: MarsPCA, TruncatedSVD: MarsTruncatedSVD}

MARS_SKLEARN_DECOMP_CALLABLES = _collect_module_callables(
    mars_decomposition, sk_decomposition, skip_members=["register_op"]
)
_install_cls_members(
    SKLEARN_DECOMP_CLS_MAP, MARS_SKLEARN_DECOMP_CALLABLES, sk_decomposition
)
attach_module_callable_docstring(PCA, sk_decomposition, sk_decomposition.PCA)
attach_module_callable_docstring(
    TruncatedSVD, sk_decomposition, sk_decomposition.TruncatedSVD
)



from ...core import OutputType
from .nlargest import DataFrameNLargest


def dataframe_nsmallest(df, n, columns, keep="first"):
    """
    Return the first n rows ordered by columns in ascending order.

    Parameters
    ----------
    df : Mars DataFrame
         Input dataframe.
    n :  int
         Number of rows to return.
    columns :  label or list of labels
         Column label(s) to order by.
    keep{‘first’, ‘last’, ‘all’}, default ‘first’
        Where there are duplicate values:
            first : prioritize the first occurrence(s)
            last : prioritize the last occurrence(s)
            all : do not drop any duplicates, even it means selecting more than n items.

    Returns
    -------
    sorted_obj : Mars DataFrame
        The first n rows ordered by the given columns in ascending order.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({
    ...     'col1': ['A', 'A', 'B', 'E', 'D', 'C'],
    ...     'col2': [2, 1, 9, 8, 7, 4],
    ...     'col3': [0, 1, 9, 4, 2, 3],
    ... })
    >>> df.execute()
        col1 col2 col3
    0   A    2    0
    1   A    1    1
    2   B    9    9
    3   E    8    2
    4   D    8    4
    5   C    4    3

    Choose the first 3 rows ordered by col2

    >>> df.nsmallest(3,"col2").execute()
        col1 col2 col3
    1   A    1    1
    0   A    2    0
    5   C    4    3
    """
    if keep not in ["last", "first", "all"]:
        raise ValueError(f'''keep must be either "first", "last" or "all"''')
    if isinstance(columns, str):
        columns = [columns]
    elif isinstance(columns, list):
        columns = columns
    else:
        raise KeyError(columns)

    op = DataFrameNLargest(
        largestOrSmallest="smallest",
        n=n,
        columns=columns,
        keep=keep,
        output_types=[OutputType.dataframe],
        gpu=df.op.is_gpu(),
    )
    nlargest_df = op(df)
    return nlargest_df


def series_nsmallest(series, n, keep="first"):
    """
    Return the smallest n elements.

    Parameters
    ----------
    df : Mars Series
         Input Series.
    n :  int
         Number of rows to return.
    keep{‘first’, ‘last’, ‘all’}, default ‘first’
        Where there are duplicate values:
            first : prioritize the first occurrence(s)
            last : prioritize the last occurrence(s)
            all : do not drop any duplicates, even it means selecting more than n items.

    Returns
    -------
    sorted_obj : Mars Series
        The n smallest values in the Series, sorted in ascreasing order.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> raw = pd.Series([8, 1, 3, 10, 5])
    >>> df.execute()
    0     8.0
    1     1.0
    2     3.0
    3     10.0
    4     5.0
    dtype: float64

    Choose the smallest 3 rows

    >>> df.nsmallest(3).execute()
    1     1.0
    2     3.0
    4     5.0
    dtype: float64
    """
    if keep not in ["last", "first", "all"]:
        raise ValueError(f'''keep must be either "first", "last" or "all"''')
    op = DataFrameNLargest(
        largestOrSmallest="smallest",
        n=n,
        keep=keep,
        output_types=[OutputType.series],
        gpu=series.op.is_gpu(),
    )
    nlargest_df = op(series)
    return nlargest_df



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import os
from typing import Any, Dict, List, Optional, Union

from .._mars.utils import no_default
from ..core.adapter import mars_new_session, session


def init(
    address: Optional[str] = None,
    init_local: bool = no_default,
    session_id: Optional[str] = None,
    timeout: Optional[float] = None,
    n_worker: int = 1,
    n_cpu: Union[int, str] = "auto",
    mem_bytes: Union[int, str] = "auto",
    cuda_devices: Union[List[int], List[List[int]], str] = "auto",
    web: Union[bool, str] = "auto",
    new: bool = True,
    storage_config: Optional[Dict] = None,
    **kwargs,
) -> None:
    """
    Init Xorbits runtime locally or connect to an Xorbits cluster.

    Parameters
    ----------
    address: str, optional
        - if None which is default, address will be "127.0.0.1", a local runtime will be initialized
        - if specify an address for creating a new local runtime, specify like ``<ip>:<port>``
        - if connect to a Xorbits cluster address, e.g. ``http://<supervisor_ip>:<supervisor_web_port>``
    init_local: bool, no default value
        Indicates if creating a new local runtime.

        - If has initialized, ``init_local`` cannot be True, it will skip creating,
        - When address is None and not initialized, ``init_local`` will be True,
        - Otherwise, if it's not specified, False will be set.
    session_id: str, optional
        Session ID, if not specified, a new ID will be auto generated.
    timeout: float
        Timeout about creating a new runtime or connecting to an existing cluster.
    n_worker: int, optional
        How many workers to start when creating a local runtime.

        .. note::

          Take effect only when ``init_local`` is True

    n_cpu: int, str
        Number of CPUs, if ``auto``, the number of cores will be specified.

        .. note::

          Take effect only when ``init_local`` is True

    mem_bytes: int, str
        Memory to use, in bytes, if ``auto``, total memory bytes will be specified.

        .. note::

          Take effect only when ``init_local`` is True

    cuda_devices: list of int, list of list
        - when ``auto`` which is default, all visible GPU devices will be used
        - When n_worker is 1, list of int can be specified, means the device indexes to use
        - When n_worker > 1, list of list can be specified for each worker.

        .. note::

          Take effect only when ``init_local`` is True

    web: bool, str
        If creating a web UI.

        .. note::

          Take effect only when ``init_local`` is True

    new: bool

        If creating a new session when connecting to an existing cluster.

        .. note::

          Take effect only when ``init_local`` is False

    storage_config: dict, optional
        storage backend and its configuration when init a new local cluster.
        Using ``shared_memory`` storage backend by default.
        Currently, support `shared_memory`` and ``mmap`` two options.

        .. note::

          Take effect only when ``init_local`` is True
    """
    default_session = session.get_default_session()
    if init_local is True and default_session is not None:
        raise ValueError(
            f"`init_local` cannot be True if has initialized,"
            f"call `shutdown` before init."
        )
    if storage_config is not None:
        backends = list(storage_config.keys())
        if len(backends) > 1 or len(backends) == 0:
            raise ValueError("Only support one storage backend.")
        backend = backends[0]
        if backend not in ["shared_memory", "mmap"]:
            raise ValueError(
                "Only support one of these storage backends: `shared_memory`, `mmap`."
            )

    if init_local is no_default:
        # if address not specified and has not initialized,
        # force to initialize a local runtime.
        # otherwise when init_local not specified, set to False
        init_local = True if (address is None and default_session is None) else False

    kw: Dict[str, Any] = dict(
        address=address,
        init_local=init_local,
        session_id=session_id,
        timeout=timeout,
        new=new,
    )
    if address is None and default_session is not None:
        # if has initialized, no need to new session
        if new:  # pragma: no branch
            session.get_default_session().as_default()
        return
    if init_local:
        kw.update(
            dict(
                n_worker=n_worker,
                n_cpu=n_cpu,
                mem_bytes=mem_bytes,
                cuda_devices=cuda_devices,
                web=web,
                storage_config=storage_config,
            )
        )
    kw.update(kwargs)
    mars_new_session(**kw)


def shutdown(**kw) -> None:
    """
    Shutdown current local runtime.
    """
    sess = session.get_default_session()
    if sess:  # pragma: no branch
        # When connecting to an existing cluster by xorbits.init,
        # stop_server will not do anything,
        # so we need to delete the session first to release resources
        sess.destroy()
        sess.stop_server(**kw)


def safe_shutdown(**kw) -> None:  # pragma: no cover
    """
    Shutdown current local runtime, and ignore all the errors.
    """
    try:
        shutdown(**kw)
    except:  # noqa: E722  # pylint: disable=bare-except
        pass


# shutdown when python process exit
atexit.register(safe_shutdown)


__all__ = [
    "init",
    "shutdown",
]



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ...._mars.core.entity import OutputType
from ...._mars.serialization.serializables import DictField, Int32Field
from ...operand import DataOperand, DataOperandMixin


class HuggingfaceRechunk(DataOperand, DataOperandMixin):
    num_chunks: int = Int32Field("num_chunks")
    chunk_index: int = Int32Field("chunk_index")
    hf_kwargs = DictField("hf_kwargs")

    def __call__(self, dataset):
        return self.new_tileable([dataset], **dataset.params)

    @classmethod
    def tile(cls, op: "HuggingfaceRechunk"):
        input_chunks = []
        for inp in op.inputs:
            input_chunks.extend(inp.chunks)
        out = op.outputs[0]

        # TODO(codingl2k1): support rechunk multi chunks.
        if len(input_chunks) == 1 and op.num_chunks != 1:
            chunks = []
            for index in range(op.num_chunks):
                chunk_op = op.copy().reset_key()
                chunk_op.chunk_index = index
                c = chunk_op.new_chunk(inputs=input_chunks, index=(index, 0))
                chunks.append(c)
            return op.copy().new_tileable(
                op.inputs,
                chunks=chunks,
                nsplits=((np.nan,) * len(chunks), (np.nan,)),
                **out.params
            )
        else:
            return op.copy().new_tileable(
                op.inputs,
                chunks=input_chunks,
                nsplits=((np.nan,) * len(input_chunks), (np.nan,)),
                **out.params
            )

    @classmethod
    def execute(cls, ctx, op: "HuggingfaceRechunk"):
        inp = ctx[op.inputs[0].key]
        out_key = op.outputs[0].key
        # Default split dataset by contiguous == True.
        hf_kwargs = {"contiguous": True}
        hf_kwargs.update(op.hf_kwargs)
        ctx[out_key] = inp.shard(op.num_chunks, op.chunk_index, **hf_kwargs)


def rechunk(dataset, num_chunks: int, **hf_kwargs):
    op = HuggingfaceRechunk(
        output_types=[OutputType.huggingface_dataset],
        num_chunks=num_chunks,
        hf_kwargs=hf_kwargs,
    )
    return op(dataset)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Set, Type, Union

import pandas as pd

from .....core import TileableData
from .....core.graph import EntityGraph
from .....dataframe.core import BaseDataFrameData, BaseSeriesData, parse_index
from .....dataframe.datasource.core import ColumnPruneSupportedDataSourceMixin
from .....dataframe.groupby.aggregation import DataFrameGroupByAgg
from .....dataframe.indexing.getitem import DataFrameIndex
from .....dataframe.merge import DataFrameMerge
from .....utils import implements
from ...core import (
    OptimizationRecord,
    OptimizationRecords,
    OptimizationRecordType,
    OptimizationRule,
    Optimizer,
)
from ..core import register_optimization_rule
from .input_column_selector import InputColumnSelector
from .self_column_selector import SelfColumnSelector

OPTIMIZABLE_OP_TYPES = (DataFrameMerge, DataFrameGroupByAgg)


@register_optimization_rule()
class ColumnPruningRule(OptimizationRule):
    def __init__(
        self,
        graph: EntityGraph,
        records: OptimizationRecords,
        optimizer_cls: Type["Optimizer"],
    ):
        super().__init__(graph, records, optimizer_cls)
        self._context: Dict[TileableData, Dict[TileableData, Set[Any]]] = {}

    def _get_successor_required_columns(self, data: TileableData) -> Set[Any]:
        """
        Get columns required by the successors of the given tileable data.
        """
        successors = self._get_successors(data)
        if successors:
            res = set().union(
                *[self._context[successor][data] for successor in successors]
            )
            # When getting the required columns of a DataFrameIndex node, we need to consider itself.
            if (
                isinstance(data, BaseDataFrameData)
                and isinstance(data.op, DataFrameIndex)
                and len(data.dtypes) > 0
            ):
                res = res.union(set(data.dtypes.index))
            return res
        else:
            return self._get_all_columns(data)

    @staticmethod
    def _get_self_required_columns(data: TileableData) -> Set[Any]:
        return SelfColumnSelector.select(data)

    def _get_required_columns(self, data: TileableData) -> Optional[Set[Any]]:
        required_columns = set()
        successor_required_columns = self._get_successor_required_columns(data)
        if successor_required_columns is None:
            return None
        required_columns.update(successor_required_columns)
        self_required_columns = self._get_self_required_columns(data)
        required_columns.update(self_required_columns)
        return required_columns

    @staticmethod
    def _get_all_columns(data: TileableData) -> Union[Set[Any], None]:
        """
        Return all the columns of given tileable data. If the given tileable data is neither
        BaseDataFrameData nor BaseSeriesData, None will be returned, indicating that column pruning
        is not available for the given tileable data.
        """
        if isinstance(data, BaseDataFrameData) and data.dtypes is not None:
            return set(data.dtypes.index)
        elif isinstance(data, BaseSeriesData):
            return {data.name}
        else:
            return None

    def _get_successors(self, data: TileableData) -> List[TileableData]:
        """
        Get successors of the given tileable data.

        Column pruning is available only when every successor is available for column pruning
        (i.e. appears in the context).
        """
        successors = list(self._graph.successors(data))
        if all(successor in self._context for successor in successors):
            return successors
        else:
            return []

    def _build_context(self) -> None:
        """
        Select required columns for each tileable data in the graph.
        """
        for data in self._graph.topological_iter(reverse=True):
            if self._is_skipped_type(data):
                continue
            self._context[data] = InputColumnSelector.select(
                data, self._get_successor_required_columns(data)
            )

    def _prune_columns(self) -> List[TileableData]:
        pruned_nodes: List[TileableData] = []
        datasource_nodes: List[TileableData] = []

        node_list = list(self._graph.topological_iter())
        for data in node_list:
            if self._is_skipped_type(data):
                continue

            op = data.op

            successor_required_columns = self._get_successor_required_columns(data)
            if (
                isinstance(op, ColumnPruneSupportedDataSourceMixin)
                and successor_required_columns is not None
                and set(successor_required_columns) != self._get_all_columns(data)
            ):
                op.set_pruned_columns(list(successor_required_columns))
                self.effective = True
                pruned_nodes.append(data)
                datasource_nodes.append(data)
                continue

            if isinstance(op, OPTIMIZABLE_OP_TYPES):
                predecessors = list(self._graph.predecessors(data))
                for predecessor in predecessors:
                    if (
                        self._is_skipped_type(predecessor)
                        or predecessor in datasource_nodes
                        # if the group by key is a series, no need to do column pruning
                        or isinstance(predecessor, BaseSeriesData)
                    ):
                        continue

                    pruned_columns = list(self._context[data][predecessor])
                    if set(pruned_columns) == self._get_all_columns(predecessor):
                        continue

                    # new node init
                    new_node_op = DataFrameIndex(
                        col_names=pruned_columns,
                    )
                    new_params = predecessor.params.copy()
                    new_params["shape"] = (
                        new_params["shape"][0],
                        len(pruned_columns),
                    )
                    new_params["dtypes"] = new_params["dtypes"][pruned_columns]
                    new_params["columns_value"] = parse_index(
                        new_params["dtypes"].index, store_data=True
                    )
                    new_node = new_node_op.new_dataframe(
                        [predecessor], **new_params
                    ).data

                    # update context
                    del self._context[data][predecessor]
                    self._context[new_node] = {predecessor: set(pruned_columns)}
                    self._context[data][new_node] = set(pruned_columns)

                    # change edges and nodes
                    self._graph.remove_edge(predecessor, data)
                    self._graph.add_node(new_node)
                    self._graph.add_edge(predecessor, new_node)
                    self._graph.add_edge(new_node, data)

                    self._records.append_record(
                        OptimizationRecord(
                            predecessor, new_node, OptimizationRecordType.new
                        )
                    )
                    # update inputs
                    data.inputs[data.inputs.index(predecessor)] = new_node
                    self.effective = True
                    pruned_nodes.extend([predecessor])
        return pruned_nodes

    def _update_tileable_params(self, pruned_nodes: List[TileableData]) -> None:
        # change dtypes and columns_value
        queue = [n for n in pruned_nodes]
        affected_nodes = set()
        while len(queue) > 0:
            node = queue.pop(0)
            if isinstance(node.op, ColumnPruneSupportedDataSourceMixin):
                affected_nodes.add(node)
            for successor in self._graph.successors(node):
                if successor not in affected_nodes:
                    queue.append(successor)
                    if not self._is_skipped_type(successor):
                        affected_nodes.add(successor)

        for node in affected_nodes:
            required_columns = self._get_required_columns(node)
            if (
                isinstance(node, BaseDataFrameData)
                and required_columns is not None
                and set(required_columns) != set(node.dtypes.index)
            ):
                new_dtypes = pd.Series(
                    dict(
                        (col, dtype)
                        for col, dtype in node.dtypes.items()
                        if col in required_columns
                    )
                )
                new_columns_value = parse_index(new_dtypes.index, store_data=True)
                node._dtypes = new_dtypes
                node._columns_value = new_columns_value
                node._shape = (node.shape[0], len(new_dtypes))

    @implements(OptimizationRule.apply)
    def apply(self):
        self._build_context()
        pruned_nodes = self._prune_columns()
        self._update_tileable_params(pruned_nodes)

    @staticmethod
    def _is_skipped_type(data: TileableData) -> bool:
        """
        If column pruning should be applied to the given tileable data.
        """
        return not isinstance(data, (BaseSeriesData, BaseDataFrameData))



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xoscar as mo


def create_actor_pool(*args, **kwargs):
    from . import dataframe, learn, remote, tensor

    modules = kwargs.pop("modules", None)
    modules = list(modules or []) + [
        tensor.__name__,
        dataframe.__name__,
        learn.__name__,
        remote.__name__,
    ]
    kwargs["modules"] = modules
    return mo.create_actor_pool(*args, **kwargs)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Optional

import xoscar as mo

from ...constants import DEFAULT_MARS_LOG_FILE_NAME, MARS_LOG_DIR_KEY

logger = logging.getLogger(__name__)


class FileLoggerActor(mo.Actor):
    """
    Read log file path from env (source from yaml config) for each node (including supervisor and all the workers).
    Expose interface for web frontend to fetch log content.
    """

    def __init__(self):
        super().__init__()
        log_dir = os.environ.get(MARS_LOG_DIR_KEY)
        self._log_file_path: Optional[str] = None
        if log_dir is None:
            # normally, this env var is set when a supervisor or worker pool is created, but some
            # test cases just use an actor pool for convenience, in which case the env var
            # MARS_LOG_DIR is None.
            logger.warning("Log directory is not set")
        else:
            self._log_file_path = os.path.join(log_dir, DEFAULT_MARS_LOG_FILE_NAME)

    def fetch_logs(self, size: int, offset: int) -> str:
        """
        Externally exposed interface.

        Parameters
        ----------
        size
        offset

        Returns
        -------

        """
        if self._log_file_path is None or not os.path.exists(self._log_file_path):
            return ""

        if size != -1:
            content = self._get_n_bytes_tail_file(size)
        else:
            content = self._get_n_bytes_from_pos(10 * 1024 * 1024, offset)
        return content

    def _get_n_bytes_tail_file(self, bytes_num: int) -> str:
        """
        Read last n bytes of file.

        Parameters
        ----------
        bytes_num: the bytes to read. -1 means read the whole file.

        Returns
        -------

        """
        f_size = os.stat(self._log_file_path).st_size
        target = f_size - bytes_num if f_size > bytes_num else 0
        with open(self._log_file_path) as f:
            f.seek(target)
            if target == 0:
                res = f.read()
            else:
                f.readline()
                res = f.read()

        return res

    def _get_n_bytes_from_pos(self, size: int, offset: int) -> str:
        """
        Read n bytes from a position.
        Parameters
        ----------
        size
        offset

        Returns
        -------

        """
        with open(self._log_file_path) as f:
            f.seek(offset)
            res = f.read(size)
        return res



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Dict

"""
An example to read csv from s3
------------------------------
>>> import mars
>>> import mars.dataframe as md
>>>
>>> mars.new_session()
>>> # Pass endpoint_url / aws_access_key_id / aws_secret_access_key to read_csv.
>>> mdf = md.read_csv("s3://bucket/example.csv", index_col=0, storage_options={
>>>     "client_kwargs": {
>>>         "endpoint_url": "http://192.168.1.12:9000",
>>>         "aws_access_key_id": "<s3 access id>",
>>>         "aws_secret_access_key": "<s3 access key>",
>>>         "aws_session_token": "<s3 session token>",
>>>     }})
>>> # Export environment vars AWS_ENDPOINT_URL / AWS_ACCESS_KEY_ID / AWS_SECRET_ACCESS_KEY / AWS_SESSION_TOKEN.
>>> mdf = md.read_csv("s3://bucket/example.csv", index_col=0)
>>> r = mdf.head(1000).execute()
>>> print(r)
"""

try:  # pragma: no cover
    # make sure s3fs is installed
    from s3fs import S3FileSystem as _S3FileSystem

    # make sure fsspec is installed
    from .fsspec_adapter import FsSpecAdapter

    del _S3FileSystem
except ImportError:
    FsSpecAdapter = None

if FsSpecAdapter is not None:  # pragma: no cover
    from .core import register_filesystem

    class S3FileSystem(FsSpecAdapter):
        def __init__(self, **kwargs):
            super().__init__("s3", **kwargs)

        @staticmethod
        def parse_from_path(uri: str):
            client_kwargs = {
                "endpoint_url": os.environ.get("AWS_ENDPOINT_URL"),
                "aws_access_key_id": os.environ.get("AWS_ACCESS_KEY_ID"),
                "aws_secret_access_key": os.environ.get("AWS_SECRET_ACCESS_KEY"),
                "aws_session_token": os.environ.get("AWS_SESSION_TOKEN"),
            }
            client_kwargs = {k: v for k, v in client_kwargs.items() if v is not None}
            return {"client_kwargs": client_kwargs}

        @classmethod
        def get_storage_options(cls, storage_options: Dict, uri: str) -> Dict:
            options = cls.parse_from_path(uri)
            for k, v in storage_options.items():
                if k == "client_kwargs":
                    options["client_kwargs"].update(v)
                else:
                    options[k] = v
            return options

    register_filesystem("s3", S3FileSystem)
else:
    S3FileSystem = None



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os

from ...lib.filesystem import LocalFileSystem, get_fs


def convert_to_abspath(
    path: str | tuple[str] | list[str], storage_options: dict | None = None
) -> str | list[str]:
    # convert path to abs_path
    if isinstance(path, (list, tuple)):
        abs_path = [
            os.path.abspath(p)
            if isinstance(get_fs(p, storage_options), LocalFileSystem)
            else p
            for p in path
        ]
    elif isinstance(get_fs(path, storage_options), LocalFileSystem):
        abs_path = os.path.abspath(path)
    else:
        abs_path = path
    return abs_path



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
import heapq
import json
import logging
import operator
from collections import Counter

from xoscar.profiling import _CallStats
from xoscar.profiling import _ProfilingData as XoscarProfilingData
from xoscar.profiling import _ProfilingOptionDescriptor
from xoscar.profiling import _ProfilingOptions as XoscarProfilingOptions

from .typing import BandType

logger = logging.getLogger(__name__)


class _ProfilingOptions(XoscarProfilingOptions):
    debug_interval_seconds = _ProfilingOptionDescriptor(float, default=None)
    slow_calls_duration_threshold = _ProfilingOptionDescriptor(int, default=1)
    slow_subtasks_duration_threshold = _ProfilingOptionDescriptor(int, default=10)


class _SubtaskStats:
    def __init__(self, options: _ProfilingOptions):
        self._options = options
        self._band_counter = Counter()
        self._slow_subtasks = []

    def collect(self, subtask, band: BandType, duration: float):
        band_address = band[0]
        self._band_counter[band_address] += 1
        if duration < self._options.slow_subtasks_duration_threshold:
            return
        key = (duration, band_address, subtask)
        try:
            if len(self._slow_subtasks) < 10:
                heapq.heappush(self._slow_subtasks, key)
            else:
                heapq.heapreplace(self._slow_subtasks, key)
        except TypeError:
            pass

    def to_dict(self) -> dict:
        band_subtasks = {}
        key = operator.itemgetter(1)
        if len(self._band_counter) > 10:
            items = self._band_counter.items()
            band_subtasks.update(heapq.nlargest(5, items, key=key))
            band_subtasks.update(reversed(heapq.nsmallest(5, items, key=key)))
        else:
            band_subtasks.update(
                sorted(self._band_counter.items(), key=key, reverse=True)
            )
        slow_subtasks = {}
        for duration, band, subtask in sorted(
            self._slow_subtasks, key=operator.itemgetter(0), reverse=True
        ):
            slow_subtasks[f"[{band}]{subtask}"] = duration
        return {"band_subtasks": band_subtasks, "slow_subtasks": slow_subtasks}


class _ProfilingData(XoscarProfilingData):
    def __init__(self):
        super().__init__()
        self._subtask_stats = {}

    def init(self, task_id: str, options=None):
        options = _ProfilingOptions(options)
        logger.info(
            "Init profiling data for task %s with debug interval seconds %s.",
            task_id,
            options.debug_interval_seconds,
        )
        self._data[task_id] = {
            "general": {},
            "serialization": {},
            "most_calls": {},
            "slow_calls": {},
            "band_subtasks": {},
            "slow_subtasks": {},
        }
        self._call_stats[task_id] = _CallStats(options)
        self._subtask_stats[task_id] = _SubtaskStats(options)

        async def _debug_profiling_log():
            while True:
                try:
                    r = self._data.get(task_id, None)
                    if r is None:
                        logger.info("Profiling debug log break.")
                        break
                    r = copy.copy(r)  # shadow copy is enough.
                    r.update(self._call_stats.get(task_id).to_dict())
                    r.update(self._subtask_stats.get(task_id).to_dict())
                    logger.warning("Profiling debug:\n%s", json.dumps(r, indent=4))
                except Exception:
                    logger.exception("Profiling debug log failed.")
                await asyncio.sleep(options.debug_interval_seconds)

        if options.debug_interval_seconds is not None:
            self._debug_task[task_id] = task = asyncio.create_task(
                _debug_profiling_log()
            )
            task.add_done_callback(lambda _: self._debug_task.pop(task_id, None))

    def pop(self, task_id: str):
        logger.info("Pop profiling data of task %s.", task_id)
        debug_task = self._debug_task.pop(task_id, None)
        if debug_task is not None:
            debug_task.cancel()
        r = self._data.pop(task_id, None)
        if r is not None:
            r.update(self._call_stats.pop(task_id).to_dict())
            r.update(self._subtask_stats.pop(task_id).to_dict())
        return r

    def collect_subtask(self, subtask, band: BandType, duration: float):
        if self._subtask_stats:
            stats = self._subtask_stats.get(subtask.task_id)
            if stats is not None:
                stats.collect(subtask, band, duration)


XoscarProfilingData.set_instance(_ProfilingData())
ProfilingData = XoscarProfilingData.get_instance()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
from collections import defaultdict, deque
from typing import Dict, List, Tuple, Type, Union

from ....config import Config
from ....core import ChunkGraph, ChunkType, enter_mode
from ....core.operand import (
    Fetch,
    LogicKeyGenerator,
    MapReduceOperand,
    OperandStage,
    ShuffleFetchType,
    ShuffleProxy,
    VirtualOperand,
)
from ....lib.ordered_set import OrderedSet
from ....resource import Resource
from ....typing import BandType, OperandType
from ....utils import build_fetch, build_fetch_shuffle, tokenize
from ...subtask import Subtask, SubtaskGraph
from ..core import MapReduceInfo, Task, new_task_id
from .assigner import AbstractGraphAssigner, GraphAssigner
from .fusion import Coloring

logger = logging.getLogger(__name__)


def need_reassign_worker(op: OperandType) -> bool:
    # NOTE(qinxuye): special process for reducer
    # We'd better set reducer op's stage to reduce, however,
    # in many case, we copy a reducer op from tileable op,
    # then set stage as reducer one,
    # it would be quite nasty to take over the __setattr__ and
    # make reassign_worker True etc.
    return op.reassign_worker or (
        isinstance(op, MapReduceOperand) and op.stage == OperandStage.reduce
    )


class GraphAnalyzer:
    """
    An subtask graph builder which build subtask graph for chunk graph based on passed band_resource.

    If push shuffle is used, this builder will validate predecessors orders consistency of ShuffleProxy between
    chunk graph and generated subtask graph.
    """

    _map_reduce_id = itertools.count()

    def __init__(
        self,
        chunk_graph: ChunkGraph,
        band_resource: Dict[BandType, Resource],
        task: Task,
        config: Config,
        chunk_to_subtasks: Dict[ChunkType, Subtask],
        graph_assigner_cls: Type[AbstractGraphAssigner] = None,
        stage_id: str = None,
        map_reduce_id_to_infos: Dict[int, MapReduceInfo] = None,
        shuffle_fetch_type: ShuffleFetchType = ShuffleFetchType.FETCH_BY_KEY,
    ):
        self._chunk_graph = chunk_graph
        self._final_result_chunks_set = set(self._chunk_graph.result_chunks)
        self._band_resource = band_resource
        self._task = task
        self._stage_id = stage_id
        self._config = config
        self._shuffle_fetch_type = shuffle_fetch_type
        self._has_shuffle = any(
            isinstance(c.op, MapReduceOperand) for c in self._chunk_graph
        )
        self._fuse_enabled = task.fuse_enabled
        self._extra_config = task.extra_config
        self._chunk_to_subtasks = chunk_to_subtasks
        self._map_reduce_id_to_infos = map_reduce_id_to_infos
        if graph_assigner_cls is None:
            graph_assigner_cls = GraphAssigner
        self._graph_assigner_cls = graph_assigner_cls
        self._chunk_to_copied = dict()
        self._logic_key_generator = LogicKeyGenerator()

    @classmethod
    def next_map_reduce_id(cls) -> int:
        return next(cls._map_reduce_id)

    @classmethod
    def _iter_start_ops(cls, chunk_graph: ChunkGraph):
        visited = set()
        op_keys = set()
        start_chunks = deque(chunk_graph.iter_indep())
        stack = deque([start_chunks.popleft()])

        while stack:
            chunk = stack.popleft()
            if chunk not in visited:
                inp_chunks = chunk_graph.predecessors(chunk)
                if not inp_chunks or all(
                    inp_chunk in visited for inp_chunk in inp_chunks
                ):
                    if len(inp_chunks) == 0:
                        op_key = chunk.op.key
                        if op_key not in op_keys:
                            op_keys.add(op_key)
                            yield chunk.op
                    visited.add(chunk)
                    stack.extend(c for c in chunk_graph[chunk] if c not in visited)
                else:
                    stack.appendleft(chunk)
                    stack.extendleft(
                        reversed(
                            [
                                c
                                for c in chunk_graph.predecessors(chunk)
                                if c not in visited
                            ]
                        )
                    )
            if not stack and start_chunks:
                stack.appendleft(start_chunks.popleft())

    def _gen_input_chunks(
        self,
        inp_chunks: List[ChunkType],
        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType],
    ) -> List[ChunkType]:
        # gen fetch chunks for input chunks
        inp_fetch_chunks = []
        for inp_chunk in inp_chunks:
            if inp_chunk in chunk_to_fetch_chunk:
                inp_fetch_chunks.append(chunk_to_fetch_chunk[inp_chunk])
            elif isinstance(inp_chunk.op, Fetch):
                chunk_to_fetch_chunk[inp_chunk] = inp_chunk
                inp_fetch_chunks.append(inp_chunk)
            elif isinstance(inp_chunk.op, ShuffleProxy):
                n_reducers = inp_chunk.op.n_reducers
                fetch_chunk = build_fetch_shuffle(
                    inp_chunk,
                    n_reducers=n_reducers,
                    shuffle_fetch_type=self._shuffle_fetch_type,
                ).data
                chunk_to_fetch_chunk[inp_chunk] = fetch_chunk
                inp_fetch_chunks.append(fetch_chunk)
            else:
                fetch_chunk = build_fetch(inp_chunk).data
                chunk_to_fetch_chunk[inp_chunk] = fetch_chunk
                inp_fetch_chunks.append(fetch_chunk)

        return inp_fetch_chunks

    @staticmethod
    def _to_band(band_or_worker: Union[BandType, str]) -> BandType:
        if isinstance(band_or_worker, tuple) and len(band_or_worker) == 2:
            # band already
            return band_or_worker
        else:
            return band_or_worker, "numa-0"

    @staticmethod
    def _get_expect_band(op: OperandType):
        if op.expect_band is not None:
            return op.expect_band
        elif op.expect_worker is not None:
            return GraphAnalyzer._to_band(op.expect_worker)

    def _gen_subtask_info(
        self,
        chunks: List[ChunkType],
        chunk_to_subtask: Dict[ChunkType, Subtask],
        chunk_to_bands: Dict[ChunkType, BandType],
        chunk_to_fetch_chunk: Dict[ChunkType, ChunkType],
    ) -> Tuple[Subtask, List[Subtask], bool]:
        # gen subtask and its input subtasks
        chunks_set = set(chunks)
        result_chunks = []
        result_chunks_set = set()
        chunk_graph = ChunkGraph(result_chunks)
        out_of_scope_chunks = []
        chunk_to_copied = self._chunk_to_copied
        update_meta_chunks = []
        # subtask properties
        band = None
        is_virtual = None
        retryable = True
        chunk_priority = None
        expect_band = None
        bands_specified = None
        processed = set()
        for chunk in chunks:
            if chunk in processed:
                continue
            if expect_band is None:
                expect_band = self._get_expect_band(chunk.op)
                bands_specified = expect_band is not None
            else:  # pragma: no cover
                curr_expect_band = self._get_expect_band(chunk.op)
                assert curr_expect_band is None or expect_band == curr_expect_band, (
                    f"expect_band {curr_expect_band} conflicts with chunks that have same color: "
                    f"{expect_band}"
                )
            # process band
            chunk_band = chunk_to_bands.get(chunk)
            if chunk_band is not None:
                assert (
                    band is None or band == chunk_band
                ), "band conflicts with chunks that have same color"
                band = chunk_band
            # process is_virtual
            if isinstance(chunk.op, VirtualOperand):
                assert is_virtual is None, "only 1 virtual operand can exist"
                is_virtual = True
            else:
                is_virtual = False
            # process retryable
            if not chunk.op.retryable:
                retryable = False
            # process priority
            if chunk.op.priority is not None:
                assert (
                    chunk_priority is None or chunk_priority == chunk.op.priority
                ), "priority conflicts with chunks that have same color"
                chunk_priority = chunk.op.priority
            # process input chunks
            inp_chunks = []
            build_fetch_index_to_chunks = dict()
            for i, inp_chunk in enumerate(chunk.inputs):
                if inp_chunk in chunks_set:
                    inp_chunks.append(chunk_to_copied[inp_chunk])
                else:
                    build_fetch_index_to_chunks[i] = inp_chunk
                    inp_chunks.append(None)
                    if not isinstance(inp_chunk.op, Fetch):
                        out_of_scope_chunks.append(inp_chunk)
            fetch_chunks = self._gen_input_chunks(
                list(build_fetch_index_to_chunks.values()), chunk_to_fetch_chunk
            )
            for i, fetch_chunk in zip(build_fetch_index_to_chunks, fetch_chunks):
                inp_chunks[i] = fetch_chunk
            copied_op = chunk.op.copy()
            copied_op._key = chunk.op.key
            out_chunks = [
                c.data
                for c in copied_op.new_chunks(
                    inp_chunks, kws=[c.params.copy() for c in chunk.op.outputs]
                )
            ]
            for src_chunk, out_chunk in zip(chunk.op.outputs, out_chunks):
                processed.add(src_chunk)
                out_chunk._key = src_chunk.key
                chunk_graph.add_node(out_chunk)
                # cannot be copied twice
                assert src_chunk not in chunk_to_copied
                chunk_to_copied[src_chunk] = out_chunk
                if src_chunk in self._final_result_chunks_set:
                    if out_chunk not in result_chunks_set:
                        # add to result chunks
                        result_chunks.append(out_chunk)
                        # chunk is in the result chunks of full chunk graph
                        # meta need to be updated
                        update_meta_chunks.append(out_chunk)
                        result_chunks_set.add(out_chunk)
                if not is_virtual:
                    # skip adding fetch chunk to chunk graph when op is virtual operand
                    for c in inp_chunks:
                        if c not in chunk_graph:
                            chunk_graph.add_node(c)
                        chunk_graph.add_edge(c, out_chunk)
        stage_n_outputs = len(result_chunks)
        # add chunks with no successors into result chunks
        result_chunks.extend(
            c
            for c in chunk_graph.iter_indep(reverse=True)
            if c not in result_chunks_set
        )
        expect_bands = (
            [expect_band] if bands_specified else ([band] if band is not None else None)
        )
        # calculate priority
        if out_of_scope_chunks:
            inp_subtasks = []
            for out_of_scope_chunk in out_of_scope_chunks:
                copied_out_of_scope_chunk = chunk_to_copied[out_of_scope_chunk]
                inp_subtask = chunk_to_subtask[out_of_scope_chunk]
                if (
                    copied_out_of_scope_chunk
                    not in inp_subtask.chunk_graph.result_chunks
                ):
                    # make sure the chunk that out of scope
                    # is in the input subtask's results,
                    # or the meta may be lost
                    inp_subtask.chunk_graph.result_chunks.append(
                        copied_out_of_scope_chunk
                    )
                inp_subtasks.append(inp_subtask)
            depth = max(st.priority[0] for st in inp_subtasks) + 1
        else:
            inp_subtasks = []
            depth = 0
        priority = (depth, chunk_priority or 0)

        subtask = Subtask(
            subtask_id=new_task_id(),
            stage_id=self._stage_id,
            logic_key=self._gen_logic_key(chunks),
            session_id=self._task.session_id,
            task_id=self._task.task_id,
            chunk_graph=chunk_graph,
            expect_bands=expect_bands,
            bands_specified=bands_specified,
            virtual=is_virtual,
            priority=priority,
            retryable=retryable,
            update_meta_chunks=update_meta_chunks,
            extra_config=self._extra_config,
            stage_n_outputs=stage_n_outputs,
        )

        is_shuffle_proxy = False
        if self._has_shuffle:
            proxy_chunks = [c for c in result_chunks if isinstance(c.op, ShuffleProxy)]
            if proxy_chunks:
                assert len(proxy_chunks) <= 1, proxy_chunks
                is_shuffle_proxy = True
        return subtask, inp_subtasks, is_shuffle_proxy

    def _gen_logic_key(self, chunks: List[ChunkType]):
        return tokenize(
            *[self._logic_key_generator.get_logic_key(chunk.op) for chunk in chunks]
        )

    def _gen_map_reduce_info(
        self, chunk: ChunkType, assign_results: Dict[ChunkType, BandType]
    ):
        reducer_ops = OrderedSet(
            [
                c.op
                for c in self._chunk_graph.successors(chunk)
                if c.op.stage == OperandStage.reduce
            ]
        )
        map_chunks = [
            c
            for c in self._chunk_graph.predecessors(chunk)
            if (c.op.stage == OperandStage.map) or c.is_mapper
        ]
        map_reduce_id = self.next_map_reduce_id()
        for map_chunk in map_chunks:
            # record analyzer map reduce id for mapper op
            # copied chunk exists because map chunk must have
            # been processed before shuffle proxy
            copied_map_chunk = self._chunk_to_copied[map_chunk]
            if not hasattr(copied_map_chunk, "extra_params"):  # pragma: no cover
                copied_map_chunk.extra_params = dict()
            copied_map_chunk.extra_params["analyzer_map_reduce_id"] = map_reduce_id
        reducer_bands = [assign_results[r.outputs[0]] for r in reducer_ops]
        map_reduce_info = MapReduceInfo(
            map_reduce_id=map_reduce_id,
            reducer_indexes=[reducer_op.reducer_index for reducer_op in reducer_ops],
            reducer_bands=reducer_bands,
        )
        self._map_reduce_id_to_infos[map_reduce_id] = map_reduce_info

    @enter_mode(build=True)
    def gen_subtask_graph(
        self, op_to_bands: Dict[str, BandType] = None
    ) -> SubtaskGraph:
        """
        Analyze chunk graph and generate subtask graph.

        Returns
        -------
        subtask_graph: SubtaskGraph
            Subtask graph.
        """
        # reassign worker when specified reassign_worker = True
        # or it's a reducer operands
        reassign_worker_ops = [
            chunk.op for chunk in self._chunk_graph if need_reassign_worker(chunk.op)
        ]
        start_ops = (
            list(self._iter_start_ops(self._chunk_graph))
            if len(self._chunk_graph) > 0
            else []
        )

        # assign start chunks
        to_assign_ops = start_ops + reassign_worker_ops
        assigner = self._graph_assigner_cls(
            self._chunk_graph, to_assign_ops, self._band_resource
        )
        # assign expect bands
        cur_assigns = {
            op.key: self._get_expect_band(op)
            for op in start_ops
            if op.expect_band is not None or op.expect_worker is not None
        }
        if op_to_bands:
            cur_assigns.update(op_to_bands)
        logger.debug(
            "Start to assign %s start chunks for task %s",
            len(start_ops),
            self._task.task_id,
        )
        chunk_to_bands = assigner.assign(cur_assigns=cur_assigns)
        logger.debug(
            "Assigned %s start chunks for task %s", len(start_ops), self._task.task_id
        )
        # assign expect workers for those specified with `expect_worker` or `expect_band`
        # skip `start_ops`, which have been assigned before
        start_ops_set = set(start_ops)
        for chunk in self._chunk_graph:
            if chunk not in start_ops_set:
                if chunk.op.expect_band is not None:
                    chunk_to_bands[chunk] = chunk.op.expect_band
                elif chunk.op.expect_worker is not None:
                    chunk_to_bands[chunk] = self._to_band(chunk.op.expect_worker)

        # color nodes
        if self._fuse_enabled:
            logger.debug("Start to fuse chunks for task %s", self._task.task_id)
            # sort start chunks in coloring as start_ops
            op_key_to_chunks = defaultdict(list)
            for chunk in self._chunk_graph:
                op_key_to_chunks[chunk.op.key].append(chunk)
            init_chunk_to_bands = dict()
            for start_op in start_ops:
                for start_chunk in op_key_to_chunks[start_op.key]:
                    init_chunk_to_bands[start_chunk] = chunk_to_bands[start_chunk]
            if (
                self._has_shuffle
                and self._shuffle_fetch_type == ShuffleFetchType.FETCH_BY_INDEX
            ):
                # ensure no shuffle mapper chunks fused into same subtask.
                initial_same_color_num = 1
            else:
                initial_same_color_num = getattr(
                    self._config, "initial_same_color_num", None
                )
            coloring = Coloring(
                self._chunk_graph,
                self._band_resource,
                init_chunk_to_bands,
                initial_same_color_num=initial_same_color_num,
                as_broadcaster_successor_num=getattr(
                    self._config, "as_broadcaster_successor_num", None
                ),
            )
            chunk_to_colors = coloring.color()
        else:
            # if not fuse enabled, color all chunks with different colors
            op_to_colors = dict()
            chunk_to_colors = dict()
            color_gen = itertools.count()
            for c in self._chunk_graph.topological_iter():
                if c.op not in op_to_colors:
                    chunk_to_colors[c] = op_to_colors[c.op] = next(color_gen)
                else:
                    chunk_to_colors[c] = op_to_colors[c.op]
        color_to_chunks = defaultdict(list)
        for chunk, color in chunk_to_colors.items():
            if not isinstance(chunk.op, Fetch):
                color_to_chunks[color].append(chunk)

        # gen subtask graph
        subtask_graph = SubtaskGraph()
        chunk_to_fetch_chunk = dict()
        chunk_to_subtask = self._chunk_to_subtasks
        # states
        visited = set()
        logic_key_to_subtasks = defaultdict(list)
        if self._shuffle_fetch_type == ShuffleFetchType.FETCH_BY_INDEX:
            for chunk in self._chunk_graph.topological_iter():
                if not isinstance(chunk.op, ShuffleProxy):
                    continue
                # Can't use `OperandStage.map` to find mappers directly, since `stage` of some operand
                # such as `DataFrameIndexAlign` are `OperandStage.map` but not a shuffle mapper sometimes.
                mapper_chunks = self._chunk_graph.predecessors(chunk)
                for mapper_chunk in mapper_chunks:
                    chunk_color = chunk_to_colors[mapper_chunk]
                    same_color_chunks = color_to_chunks[chunk_color]
                    mappers = [
                        c
                        for c in same_color_chunks
                        if c.op.stage == OperandStage.map
                        and any(
                            isinstance(succ.op, ShuffleProxy)
                            for succ in self._chunk_graph.iter_successors(c)
                        )
                    ]
                    if len(mappers) > 1:
                        # ensure every subtask contains only at most one mapper
                        for mapper in mappers:
                            same_color_chunks.remove(mapper)
                            mapper_color = coloring.next_color()
                            chunk_to_colors[mapper] = mapper_color
                            color_to_chunks[mapper_color] = [mapper]
        for chunk in self._chunk_graph.topological_iter():
            if chunk in visited or isinstance(chunk.op, Fetch):
                # skip fetch chunk
                continue

            color = chunk_to_colors[chunk]
            same_color_chunks = color_to_chunks[color]
            if all(isinstance(c.op, Fetch) for c in same_color_chunks):
                # all fetch ops, no need to gen subtask
                continue
            subtask, inp_subtasks, is_shuffle_proxy = self._gen_subtask_info(
                same_color_chunks,
                chunk_to_subtask,
                chunk_to_bands,
                chunk_to_fetch_chunk,
            )
            subtask_graph.add_node(subtask)
            if is_shuffle_proxy:
                subtask_graph.add_shuffle_proxy_subtask(subtask)
            logic_key_to_subtasks[subtask.logic_key].append(subtask)
            for inp_subtask in inp_subtasks:
                subtask_graph.add_edge(inp_subtask, subtask)

            for c in same_color_chunks:
                chunk_to_subtask[c] = subtask
            if self._map_reduce_id_to_infos is not None and isinstance(
                chunk.op, ShuffleProxy
            ):
                self._gen_map_reduce_info(chunk, chunk_to_bands)
            visited.update(same_color_chunks)

        for subtasks in logic_key_to_subtasks.values():
            for logic_index, subtask in enumerate(subtasks):
                subtask.logic_index = logic_index
                subtask.logic_parallelism = len(subtasks)
        return subtask_graph
